""" simple token service """
import datetime
import uuid
import sqlahelper as psa
import sqlalchemy as sqla

from ptah import config
from sqla import QueryFreezer

__all__ = ['TokenType', 'service']

TOKEN_TYPE = 'ptah:token-type'


class TokenType(object):
    """ Token type interface

    ``id`` unique token type id. Token service uses this id
    for token type identification in tokens storage.

    ``timeout`` token timout, it has to be timedelta instance.
    """

    def __init__(self, id, timeout, title='', description=''):
        self.id = id
        self.timeout = timeout
        self.title = title
        self.description = description

        info = config.DirectiveInfo()
        info.attach(
            config.Action(
                lambda config, id, tp: \
                    config.get_cfg_storage(TOKEN_TYPE).update({id: tp}),
                (id, self),
                discriminator=(TOKEN_TYPE, id))
            )


class TokenService(object):
    """ Token management service """

    _sql_get = QueryFreezer(
        lambda: Session.query(Token).filter(
            Token.token == sqla.sql.bindparam('token')))

    _sql_get_by_data = QueryFreezer(
        lambda: Session.query(Token).filter(
            sqla.sql.and_(Token.data == sqla.sql.bindparam('data'),
                          Token.typ == sqla.sql.bindparam('typ'))))

    def generate(self, typ, data):
        """ Generate and return string token.

        ``type`` object implemented ITokenType interface.

        ``data`` token type specific data, it must be python string. """

        t = Token(typ, data)
        Session.add(t)
        Session.flush()
        return t.token

    def get(self, token):
        """ Get data for token """
        t = self._sql_get.first(token=token)
        if t is not None:
            return t.data

    def get_bydata(self, typ, data):
        """ Get token for data """
        t = self._sql_get_by_data.first(data=data, typ=typ.id)
        if t is not None:
            return t.token

    def remove(self, token):
        """ Remove token """
        Session.query(Token).filter(
            sqla.sql.or_(Token.token == token,
                         Token.valid > datetime.datetime.now())).delete()


service = TokenService()

Base = psa.get_base()
Session = psa.get_session()


class Token(Base):

    __tablename__ = 'ptah_tokens'

    id = sqla.Column(sqla.Integer, primary_key=True)
    token = sqla.Column(sqla.Unicode(48))
    valid = sqla.Column(sqla.DateTime)
    data = sqla.Column(sqla.Unicode)
    typ = sqla.Column(sqla.Unicode(48))

    def __init__(self, typ, data):
        super(Token, self).__init__()

        self.typ = typ.id
        self.data = data
        self.valid = datetime.datetime.now() + typ.timeout
        self.token = uuid.uuid4().get_hex()



from ptah import view, mail
from pyramid.testing import DummyRequest

from base import Base


class Content(object):
    pass


class TestMailTemplate(Base):

    def _make_one(self):
        from ptah.mail import MailTemplate

        class MyMailTemplate(MailTemplate):
            template = view.template('test_mail_tmpl.pt')
            subject = 'Test subject'
            charset = 'utf-8'
            content_type = 'text/html'

        return MyMailTemplate

    def test_mailtmpl_ctor(self):
        tmpl = mail.MailTemplate(Content(), DummyRequest(),
                                 testattr = 'testattr')

        self.assertEqual(tmpl.testattr, 'testattr')

    def test_mailtmpl_basics(self):
        self._init_ptah()

        cls = self._make_one()
        cls.message_id = 'message id'

        tmpl = cls(Content(), DummyRequest())()

        self.assertEqual(
            tmpl['Content-Type'], 'text/html; charset="utf-8"')
        self.assertEqual(
            tmpl['Content-Transfer-Encoding'], 'base64')
        self.assertEqual(
            tmpl['Subject'], '=?utf-8?q?Test_subject?=')
        self.assertEqual(
            tmpl['Message-ID'], 'message id')
        self.assertEqual(
            tmpl['From'], 'Site administrator <administrator@localhost.org>')

    def test_mailtmpl_from(self):
        self._init_ptah()

        cls = self._make_one()

        cls.from_name = 'Test'
        cls.from_address = 'ptah@ptahproject.com'

        tmpl = cls(Content(), DummyRequest())()
        self.assertEqual(tmpl['From'], 'Test <ptah@ptahproject.com>')

    def test_mailtmpl_to(self):
        self._init_ptah()

        cls = self._make_one()

        cls.to_address = 'ptah@ptahproject.com'

        tmpl = cls(Content(), DummyRequest())()
        self.assertEqual(tmpl['To'], 'ptah@ptahproject.com')

    def test_mailtmpl_headers(self):
        self._init_ptah()

        cls = self._make_one()

        tmpl = cls(Content(), DummyRequest())
        self.assertFalse(tmpl.has_header('X-Mailer'))

        tmpl.add_header('X-Mailer', 'ptah')
        self.assertTrue(tmpl.has_header('X-Mailer'))

        msg = tmpl()
        self.assertEqual(msg['X-Mailer'], 'ptah')

    def test_mailtmpl_headers_encoding(self):
        self._init_ptah()

        cls = self._make_one()

        tmpl = cls(Content(), DummyRequest())
        tmpl.add_header('X-Mailer', 'ptah', True)

        msg = tmpl()
        self.assertEqual(msg['X-Mailer'], '=?utf-8?q?ptah?=')

    def test_mailtmpl_headers_gen(self):
        self._init_ptah()

        cls = self._make_one()
        tmpl = cls(Content(), DummyRequest())

        msg = tmpl(**{'X-Mailer': 'ptah'})
        self.assertEqual(msg['X-Mailer'], 'ptah')

        msg = tmpl(**{'X-Mailer': ('ptah', True)})
        self.assertEqual(str(msg['X-Mailer']), '=?utf-8?q?ptah?=')

    def test_mailtmpl_attachment(self):
        self._init_ptah()

        cls = self._make_one()
        tmpl = cls(Content(), DummyRequest())
        self.assertEqual(tmpl.get_attachments(), [])

        tmpl.add_attachment('File data', 'text/plain', 'file.txt')
        self.assertEqual(
            tmpl.get_attachments(),
            [('File data', 'text/plain', 'file.txt', 'attachment')])

        msg = tmpl()
        payload = msg.get_payload()

        self.assertTrue(msg.is_multipart())
        self.assertEqual(
            payload[0]['Content-Id'], '<file.txt@ptah>')
        self.assertEqual(
            payload[0]['Content-Disposition'],
            'attachment; filename="file.txt"')

    def test_mailtmpl_attachment_inline(self):
        self._init_ptah()

        cls = self._make_one()
        tmpl = cls(Content(), DummyRequest())
        self.assertEqual(tmpl.get_attachments(), [])

        tmpl.add_attachment('File data', 'text/plain', 'file.txt', 'inline')
        self.assertEqual(
            tmpl.get_attachments(),
            [('File data', 'text/plain', 'file.txt', 'inline')])

        msg = tmpl()
        payload = msg.get_payload()[0]
        payload = payload.get_payload()[-1]

        self.assertEqual(
            payload['Content-Disposition'],
            'inline; filename="file.txt"')

    def test_mailtmpl_alternative(self):
        self._init_ptah()

        cls = self._make_one()

        tmpl = cls(Content(), DummyRequest())
        tmpl.add_header('x-tmpl', 'Template1')
        tmpl2 = cls(Content(), DummyRequest())
        tmpl2.add_header('x-tmpl', 'Template2')

        tmpl.add_alternative(tmpl2)
        self.assertEqual(tmpl.get_alternative(), [tmpl2])

        msg = tmpl()

        payload = msg.get_payload()
        self.assertEqual(msg['x-tmpl'], 'Template1')
        self.assertEqual(payload[1]['x-tmpl'], 'Template2')

    def test_mailtmpl_send(self):
        from ptah.settings import MAIL

        self._init_ptah()

        cls = self._make_one()
        tmpl = cls(Content(), DummyRequest())

        data = []
        class Mailer(object):

            def send(self, frm, to, msg):
                data.append((frm, to, msg))

        MAIL.Mailer = Mailer()

        tmpl.send()

        self.assertEqual(
            data[0][0], 'Site administrator <administrator@localhost.org>')
        self.assertEqual(
            data[0][1], None)
        self.assertIn('From: Site administrator <administrator@localhost.org>',
                      data[0][2].as_string())

        tmpl.send('test@ptahproject.org')
        self.assertEqual(data[-1][1], 'test@ptahproject.org')

        del MAIL.Mailer



import sys
import unittest
from pyramid.router import Router
from pyramid.config import Configurator

import ptah


class TestPtahInit(unittest.TestCase):

    def test_init_includeme(self):
        config = Configurator()
        config.include('ptah')

        self.assertTrue(hasattr(config, 'ptah_initialize'))

    def test_init_ptah_init(self):
        config = Configurator(
            settings = {'sqla.url': 'sqlite://', 'ptah.excludes': ''})

        data = [False, False]

        def initialized_handler(ev):
            data[0] = True

        def settings_initialized_handler(ev):
            data[1] = True

        sm = config.registry
        sm.registerHandler(
            initialized_handler,
            (ptah.config.Initialized,))
        sm.registerHandler(
            settings_initialized_handler,
            (ptah.config.SettingsInitialized,))

        config.include('ptah')
        config.ptah_initialize(autoinclude=True)

        self.assertTrue(data[0])
        self.assertTrue(data[1])

    def test_init_ptah_init_exception(self):
        config = Configurator(
            settings = {'sqla.url': 'sqlite://'})

        class CustomException(Exception):
            pass

        def initialized_handler(ev):
            raise CustomException()

        sm = config.registry
        sm.registerHandler(
            initialized_handler,
            (ptah.config.Initialized,))

        config.include('ptah')

        try:
            config.ptah_initialize()
        except Exception, err:
            pass

        self.assertIsInstance(err, ptah.config.StopException)
        self.assertIsInstance(err.exc, CustomException)

    def test_init_ptah_init_settings_exception(self):
        config = Configurator(
            settings = {'sqla.url': 'sqlite://'})

        class CustomException(Exception):
            pass

        def initialized_handler(ev):
            raise CustomException()

        sm = config.registry
        sm.registerHandler(
            initialized_handler,
            (ptah.config.SettingsInitialized,))

        config.include('ptah')

        try:
            config.ptah_initialize(autoinclude=True)
        except Exception, err:
            pass

        self.assertIsInstance(err, ptah.config.StopException)
        self.assertIsInstance(err.exc, CustomException)

    def test_init_wsgi_app(self):
        app = ptah.make_wsgi_app({'sqla.url': 'sqlite://', 'ptah.excludes': ''})
        self.assertIsInstance(app, Router)

    def test_init_wsgi_app_exception(self):
        orig_exit = sys.exit
        orig_ptah_init = ptah.ptah_initialize

        data = [False]
        def exit(status):
            data[0] = True

        def ptah_initialize(config, packages=None, autoinclude=False):
            raise ptah.config.StopException('')

        sys.exit = exit
        ptah.ptah_initialize = ptah_initialize

        app = ptah.make_wsgi_app({})

        sys.exit = orig_exit
        ptah.ptah_initialize = orig_ptah_init

        self.assertIsNone(app)
        self.assertTrue(data[0])



import pkg_resources
import StringIO
import sys
import traceback

from collections import defaultdict
from ptah.config import directives
from pyramid.threadlocal import get_current_registry
from zope.interface.interfaces import IObjectEvent

mods = set()


class StopException(Exception):
    """ Special initialization exception means stop execution """

    def __init__(self, exc=None):
        self.exc = exc
        if isinstance(exc, BaseException):
            self.isexc = True
            self.exc_type, self.exc_value, self.exc_traceback = sys.exc_info()
        else:
            self.isexc = False

    def __str__(self):
        return '\n%s' % self.print_tb()

    def print_tb(self):
        if self.isexc and self.exc_value:
            out = StringIO.StringIO()
            traceback.print_exception(
                self.exc_type, self.exc_value, self.exc_traceback, file=out)
            return out.getvalue()
        else:
            return self.exc


class AppStarting(object):
    """ ptah sends this event when application is ready to start. """
    directives.event('Application starting event')

    config = None

    def __init__(self, config):
        self.config = config


class Initialized(object):
    """ ptah sends this after ptah.config.initialize """
    directives.event('Ptah config initialized event')

    def __init__(self, registry):
        self.registry = registry


class Config(object):

    def __init__(self, registry, actions):
        self.registry = registry
        self.actions = actions
        self.storage = defaultdict(lambda: dict())
        registry.storage = self.storage
        registry.storage['actions'] = actions

        # settings
        from settings import SETTINGS_ID
        registry.storage[SETTINGS_ID] = defaultdict(lambda: dict())

    def get_cfg_storage(self, id):
        return self.storage[id]


def initialize(config, packages=None, excludes=(),
               autoinclude=False, initsettings=True):
    """ Load ptah packages, scan and execute all configuration
    directives. """
    registry = config.registry
    registry.registerHandler(objectEventNotify, (IObjectEvent,))

    def exclude_filter(modname):
        if modname in packages:
            return True
        return exclude(modname, excludes)

    # list all packages
    if autoinclude:
        if packages is None:
            packages = list_packages(excludes=excludes)
            packages.extend([mod for mod in mods if exclude_filter(mod)])
        else:
            packages = list_packages(packages, excludes=excludes)
    elif packages is None:
        packages = ()

    # scan packages and load actions
    seen = set()
    actions = []

    for pkg in packages:
        actions.extend(directives.scan(pkg, seen, exclude_filter))

    cfg = Config(registry, actions)

    # execute actions
    actions = directives.resolveConflicts(actions)

    def runaction(action, cfg):
        cfg.action = action
        action(cfg)

    for action in actions:
        config.action(action.discriminator, runaction, (action, cfg))

    config.action(None, registry.notify, (Initialized(registry),))

    if initsettings:
        import settings

        config.action(
            None, settings.initialize_settings,
            (config.registry.settings, config))


def get_cfg_storage(id, registry=None):
    if registry is None:
        registry = get_current_registry()

    try:
        storage = registry.storage
    except:
        storage = defaultdict(lambda: dict())
        registry.storage = storage

    return storage[id]


def start(cfg):
    cfg.registry.notify(AppStarting(cfg))


def exclude(modname, excludes=()):
    for n in ('.test', '.ftest'):
        if n in modname:
            return False

    for mod in excludes:
        if modname == mod or modname.startswith(mod):
            return False
    return True


def load_package(name, seen, first=True):
    packages = []

    if name in seen:
        return packages

    seen.add(name)

    try:
        dist = pkg_resources.get_distribution(name)
        for req in dist.requires():
            pkg = req.project_name
            if pkg in seen:
                continue
            packages.extend(load_package(pkg, seen, False))

        distmap = pkg_resources.get_entry_map(dist, 'ptah')
        ep = distmap.get('package')
        if ep is not None:
            if dist.has_metadata('top_level.txt'):
                packages.extend(
                    [p.strip() for p in
                     dist.get_metadata('top_level.txt').split()])
    except pkg_resources.DistributionNotFound:
        pass

    if first and name not in packages and '-' not in name:
        packages.append(name)

    return packages


def list_packages(include_packages=None, excludes=None):
    seen = set()
    packages = []

    if include_packages is not None:
        for pkg in include_packages:
            if excludes and pkg in excludes:
                continue
            packages.extend(load_package(pkg, seen))
    else:
        for dist in pkg_resources.working_set:
            pkg = dist.project_name
            if pkg in seen:
                continue
            if excludes and pkg in excludes:
                continue

            distmap = pkg_resources.get_entry_map(dist, 'ptah')
            if 'package' in distmap:
                packages.extend(load_package(pkg, seen))
            else:
                seen.add(pkg)

    return packages


def notify(*event):
    """ Send event to event listeners """
    get_current_registry().subscribers(event, None)


def objectEventNotify(event):
    get_current_registry().subscribers((event.object, event), None)


_cleanups = set()


def cleanup(handler):
    _cleanups.add(handler)
    return handler


def cleanup_system(*modIds):
    mods.clear()

    for h in _cleanups:
        h()

    for modId in modIds:
        mod = sys.modules[modId]
        if hasattr(mod, directives.ATTACH_ATTR):
            delattr(mod, directives.ATTACH_ATTR)



""" ptah.form fields """
from ptah import view, config, manage
from ptah.form.field import FIELD_ID, PREVIEW_ID


class FieldsModule(manage.PtahModule):
    """A preview and listing of all form fields in the application.
       This is useful to see what fields are available.  You may also
       interact with the field to see how it works in display mode.
    """
    title = 'Field types'
    manage.module('fields')


class FieldsView(view.View):
    view.pview(
        context = FieldsModule,
        template = view.template('ptah.manage:templates/fields.pt'))

    def update(self):
        data = []

        fields = config.get_cfg_storage(FIELD_ID)
        previews = config.get_cfg_storage(PREVIEW_ID)

        for name, cls in fields.items():
            data.append({'name': name,
                         'doc': cls.__doc__,
                         'preview': previews.get(cls)})

        data.sort()
        self.fields = data



""" ptah rest api """
import traceback
from datetime import datetime
from cStringIO import StringIO
from simplejson import dumps
from collections import OrderedDict
from pyramid.response import Response
from pyramid.authentication import parse_ticket, AuthTicket, BadTicket
from pyramid.httpexceptions import WSGIHTTPException, HTTPNotFound

from ptah import view, config

import ptah
from ptah.settings import SECURITY


REST_ID = 'ptah:rest-service'


def RestService(name, title, description=''):
    srv = Service(name, title, description)

    apidoc = ServiceAPIDoc(name)
    srv.actions['apidoc'] = Action(apidoc, 'apidoc', apidoc.title)

    def _register(cfg, srv):
        cfg.get_cfg_storage(REST_ID)[name] = srv

    info = config.DirectiveInfo()
    info.attach(
        config.Action(
            _register, (srv, ),
            discriminator=(REST_ID, name))
        )

    return srv


class Action(object):

    def __init__(self, callable, name, title):
        self.name = name
        self.title = title
        self.callable = callable
        self.description = callable.__doc__ or ''


class Service(object):

    def __init__(self, name, title, description):
        self.name = name
        self.title = title
        self.description = description
        self.actions = {}

    def action(self, name, title):
        def wrapper(func):
            self.actions[name] = Action(func, name, title)
            return func

        return wrapper

    def __call__(self, request, action, *args):
        action = self.actions.get(action)
        if action:
            return action.callable(request, *args)
        else:
            raise HTTPNotFound


class ServiceAPIDoc(object):

    title = 'API Doc'

    def __init__(self, name):
        self.srvname = name

    def __call__(self, request):
        srv = config.get_cfg_storage(REST_ID)[self.srvname]
        url = request.application_url

        info = OrderedDict(
            (('name', srv.name),
             ('__link__', '%s/' % url),
             ('title', srv.title),
             ('description', srv.description),
             ('actions', [])))

        actions = [(a.title, a.name, a.description)
                   for a in srv.actions.values()]
        actions.sort()

        for title, name, description in actions:
            info['actions'].append(
                OrderedDict(
                    (('name', name),
                     ('__link__', '%s/%s' % (url, name)),
                     ('title', title),
                     ('description', description))))

        return info


class RestLoginRoute(object):
    """ rest login route """

    def __init__(self, request):
        self.request = request


class RestApiRoute(object):
    """ rest api route """

    def __init__(self, request):
        self.request = request


def dthandler(obj):
    return obj.isoformat() if isinstance(obj, datetime) else None


class Login(object):
    view.pview(context=RestLoginRoute)

    def __init__(self, request):
        self.request = request

    def render(self):
        request = self.request

        login = request.POST.get('login', '')
        password = request.POST.get('password', '')

        credentials = {'login': login, 'password': password}
        info = ptah.authService.authenticate(credentials)
        if info.status:
            token = self.get_token(request, info.uri)
            result = {'message': '', 'auth-token': token[:-1]}
        else:
            request.response.status = 403
            result = {'message': info.message or 'authentication failed'}

        request.response.headerslist = {'Content-Type': 'application/json'}
        return '%s\n\n' % dumps(result, indent=True, default=dthandler)

    def get_token(self, request, userid):
        secret = SECURITY.secret
        remote_addr = '0.0.0.0'

        ticket = AuthTicket(
            secret,
            userid,
            remote_addr,
            tokens=(),
            user_data='',
            cookie_name='auth-token',
            secure=False)

        return ticket.cookie_value()


class Api(object):
    view.pview(context=RestApiRoute)

    def __init__(self, request):
        self.request = request

    def render(self):
        request = self.request

        # authentication by token
        token = request.environ.get('HTTP_X_AUTH_TOKEN')
        if token:
            try:
                timestamp, userid, tokens, user_data = parse_ticket(
                    SECURITY.secret,
                    '%s!' % token,
                    '0.0.0.0')
            except BadTicket:
                userid = None

            if userid:
                ptah.authService.set_userid(userid)

        # search service and action
        service = request.matchdict['service']
        subpath = request.matchdict['subpath']
        if subpath:
            action = subpath[0]
            arguments = subpath[1:]
            if ':' in action:
                action, arg = action.split(':', 1)
                arguments = (arg,) + arguments
        else:
            action = 'apidoc'
            arguments = ()

        request.environ['SCRIPT_NAME'] = '/__rest__/%s' % service
        request.response.headerslist = {'Content-Type': 'application/json'}

        # execute action for specific service
        try:
            result = config.get_cfg_storage(REST_ID)[service](
                request, action, *arguments)
        except WSGIHTTPException, exc:
            request.response.status = exc.status
            result = {'message': str(exc)}
        except Exception, exc:
            request.response.status = 500

            out = StringIO()
            traceback.print_exc(file=out)

            result = {'message': str(exc),
                      'traceback': out.getvalue()}

        if isinstance(result, Response):
            return result

        return '%s\n\n' % dumps(result, indent=True, default=dthandler)


def enable_rest_api(config):
    """ Register /__rest__/login and /__rest__/{service}/*subpath
    routes """

    config.add_route(
        'ptah-rest-login', '/__rest__/login',
        factory=RestLoginRoute, use_global_views=True)

    config.add_route(
        'ptah-rest', '/__rest__/{service}/*subpath',
        factory=RestApiRoute, use_global_views=True)



import ptah
from ptah import cms, view, form
from pyramid.httpexceptions import HTTPFound


class SharingForm(form.Form):
    view.pview(
        'sharing.html', ptah.ILocalRolesAware,
        permission = cms.ShareContent,
        template = view.template('templates/sharing.pt'))

    csrf = True
    fields = form.Fieldset(
        form.FieldFactory(
            'text',
            'term',
            title = u'Search term',
            description = 'Searches users by login and email',
            missing = u'',
            default = u''))

    users = None
    bsize = 15

    def form_content(self):
        return {'term': self.request.session.get('sharing-search-term', '')}

    def get_principal(self, id):
        return ptah.resolve(id)

    def update(self):
        super(SharingForm, self).update()

        request = self.request
        context = self.context

        self.roles = [r for r in ptah.get_roles().values() if not r.system]
        self.local_roles = local_roles = context.__local_roles__

        term = request.session.get('sharing-search-term', '')
        if term:
            self.users = list(ptah.search_principals(term))

        if 'form.buttons.save' in request.POST:
            users = []
            userdata = {}
            for attr, val in request.POST.items():
                if attr.startswith('user-'):
                    userId, roleId = attr[5:].rsplit('-',1)
                    data = userdata.setdefault(str(userId), [])
                    data.append(str(roleId))
                if attr.startswith('userid-'):
                    users.append(str(attr.split('userid-')[-1]))

            for uid in users:
                if userdata.get(uid):
                    local_roles[str(uid)] = userdata[uid]
                elif uid in local_roles:
                    del local_roles[uid]

            context.__local_roles__ = local_roles

    @form.button('Search', actype=form.AC_PRIMARY)
    def search(self):
        data, error = self.extract()

        if not data['term']:
            self.message('Please specify search term', 'warning')
            return

        self.request.session['sharing-search-term'] = data['term']
        raise HTTPFound(location = self.request.url)



""" directives """
from ptah import config
from ptah.view.customize import LayerWrapper
from ptah.view.view import unset, register_view_impl
from ptah.view.layout import register_layout_impl
from ptah.view.snippet import register_snippet_impl


def snippet(name, context=None, template=None, layer=''):
    info = config.DirectiveInfo(allowed_scope=('class',))

    discriminator = ('ptah.view:snippet', name, context, layer)

    info.attach(
        config.ClassAction(
            LayerWrapper(register_snippet_impl, discriminator),
            (name, context, template),
            discriminator = discriminator)
        )


def pview(name=u'', context=None, route=None,
          template=None, layout=unset,
          permission='__no_permission_required__', layer=''):

    info = config.DirectiveInfo(
        allowed_scope=('class', 'module', 'function call'))

    discriminator = ('ptah.view:view', name, context, route, layer)

    if info.scope in ('module', 'function call'): # function decorator
        def wrapper(factory):
            info.attach(
                config.Action(
                    LayerWrapper(register_view_impl, discriminator),
                    (factory, name, context, template, route,
                     layout, permission),
                    discriminator = discriminator)
                )
            return factory
        return wrapper
    else:
        # class decorator
        info.attach(
            config.ClassAction(
                LayerWrapper(register_view_impl, discriminator),
                (name, context, template, route, layout, permission),
                discriminator = discriminator)
            )


def layout(name='', context=None, parent='', route=None,template=None,layer=''):
    info = config.DirectiveInfo(allowed_scope=('class',))

    discriminator = ('ptah.view:layout', name, context, route, layer)

    info.attach(
        config.ClassAction(
            LayerWrapper(register_layout_impl, discriminator),
            (name, context, template, parent, route),
            discriminator = discriminator)
        )



""" routes """
from ptah import config
from pyramid.config.util import make_predicates
from pyramid.request import route_request_iface
from pyramid.urldispatch import RoutesMapper
from pyramid.interfaces import IRequest, IRouteRequest, IRoutesMapper


def register_route(name, pattern=None, factory=None, header=None,
                   traverse=None, pregenerator=None, use_global_views=False,
                   xhr=False, request_method=None,
                   path_info=None, request_param=None,
                   accept=None, custom_predicates=()):

    # these are route predicates; if they do not match, the next route
    # in the routelist will be tried
    ignored, predicates, ignored = make_predicates(
        xhr=xhr,
        request_method=request_method,
        path_info=path_info,
        request_param=request_param,
        header=header,
        accept=accept,
        traverse=traverse,
        custom=custom_predicates)

    info = config.DirectiveInfo()
    info.attach(
        config.Action(
            register_route_impl,
            (name,pattern,factory,predicates,pregenerator,use_global_views),
            discriminator = ('ptah.view:route', name),
            order = 1))


def register_route_impl(cfg, name, pattern, factory,
                        predicates, pregenerator, use_global_views):
    request_iface = cfg.registry.queryUtility(IRouteRequest, name=name)
    if request_iface is None:
        if use_global_views:
            bases = (IRequest,)
        else:
            bases = ()
        request_iface = route_request_iface(name, bases)
        cfg.registry.registerUtility(request_iface, IRouteRequest, name=name)

    mapper = cfg.registry.queryUtility(IRoutesMapper)
    if mapper is None:
        mapper = RoutesMapper()
        cfg.registry.registerUtility(mapper, IRoutesMapper)

    return mapper.connect(name, pattern, factory, predicates=predicates,
                          pregenerator=pregenerator, static=False)



from ptah import config

from base import Base


class TestUri(Base):

    def tearDown(self):
        config.cleanup_system(self.__class__.__module__)
        super(TestUri, self).tearDown()

    def test_uri_registration(self):
        import ptah

        def resolver1(uri):
            return 'Resolved1'
        def resolver2(uri):
            return 'Resolved2'

        ptah.register_uri_resolver('test1', resolver1)
        ptah.register_uri_resolver('test2', resolver2)
        self._init_ptah()

        self.assertEqual(ptah.resolve('test1:uri'), 'Resolved1')
        self.assertEqual(ptah.resolve('test2:uri'), 'Resolved2')

        self.assertEqual(ptah.resolve(None), None)
        self.assertEqual(ptah.resolve('unknown'), None)
        self.assertEqual(ptah.resolve('unknown:uri'), None)

    def test_uri_registration_decorator(self):
        import ptah

        @ptah.resolver('test1')
        def resolver1(uri):
            return 'Resolved1'

        @ptah.resolver('test2')
        def resolver2(uri):
            return 'Resolved2'

        self._init_ptah()

        self.assertEqual(ptah.resolve('test1:uri'), 'Resolved1')
        self.assertEqual(ptah.resolve('test2:uri'), 'Resolved2')

        self.assertEqual(ptah.resolve(None), None)
        self.assertEqual(ptah.resolve('unknown'), None)
        self.assertEqual(ptah.resolve('unknown:uri'), None)

    def test_uri_registration_conflicts(self):
        import ptah
        ptah.register_uri_resolver('test', None)
        ptah.register_uri_resolver('test', None)

        self.assertRaises(config.ConflictError, self._init_ptah)

    def test_uri_registration_decorator_conflicts(self):
        import ptah

        @ptah.resolver('test1')
        def resolver1(uri): # pragma: no cover
            return 'Resolved1'

        @ptah.resolver('test1')
        def resolver2(uri): # pragma: no cover
            return 'Resolved2'

        self.assertRaises(config.ConflictError, self._init_ptah)

    def test_uri_extract_type(self):
        import ptah

        self.assertEqual(ptah.extract_uri_schema('test:uri'), 'test')
        self.assertEqual(ptah.extract_uri_schema('test'), None)
        self.assertEqual(ptah.extract_uri_schema(None), None)

    def test_uri_uri_generator(self):
        import ptah

        uri = ptah.UriFactory('test')

        u1 = uri()
        u2 = uri()

        self.assertTrue(u1.startswith('test:'))
        self.assertTrue(u2.startswith('test:'))
        self.assertTrue(u1 != u2)



""" formatters """
import pytz
import colander
import translationstring
from datetime import datetime, timedelta
from pyramid.i18n import get_localizer
from pyramid.threadlocal import get_current_request

from ptah import config

_ = translationstring.TranslationStringFactory('ptah')


class FormatImpl(dict):

    def __setitem__(self, name, formatter):
        if name in self:
            raise ValueError('Formatter "%s" already registered.' % name)

        super(FormatImpl, self).__setitem__(name, formatter)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)


format = FormatImpl()


_tzs = dict((str(tz).lower(), str(tz)) for tz in pytz.all_timezones)


class Timezone(colander.SchemaType):
    """ timezone schema type """

    def serialize(self, node, appstruct):
        if appstruct is colander.null:
            return colander.null

        return str(appstruct)

    def deserialize(self, node, cstruct):
        if cstruct is colander.null or not cstruct:
            return colander.null

        try:
            v = str(cstruct)
            if v.startswith('GMT'):
                v = 'Etc/%s' % v
            try:
                return pytz.timezone(v)
            except:
                return pytz.timezone(_tzs[v.lower()])
        except:
            raise colander.Invalid(node,
                _('"${val}" is not a timezone', mapping={'val': cstruct}))


FORMAT = config.register_settings(
    'format',

    config.SchemaNode(
        Timezone(),
        name = 'timezone',
        default = pytz.timezone('US/Central'),
        title = _('Timezone'),
        description = _('Site wide timezone.')),

    config.SchemaNode(
        colander.Str(),
        name = 'date_short',
        default = u'%m/%d/%y',
        title = _(u'Date'),
        description = _(u'Date short format')),

    config.SchemaNode(
        colander.Str(),
        name = 'date_medium',
        default = u'%b %d, %Y',
        title = _(u'Date'),
        description = _(u'Date medium format')),

    config.SchemaNode(
        colander.Str(),
        name = 'date_long',
        default = u'%B %d, %Y',
        title = _(u'Date'),
        description = _(u'Date long format')),

    config.SchemaNode(
        colander.Str(),
        name = 'date_full',
        default = u'%A, %B %d, %Y',
        title = _(u'Date'),
        description = _(u'Date full format')),

    config.SchemaNode(
        colander.Str(),
        name = 'time_short',
        default = u'%I:%M %p',
        title = _(u'Time'),
        description = _(u'Time short format')),

    config.SchemaNode(
        colander.Str(),
        name = 'time_medium',
        default = u'%I:%M %p',
        title = _(u'Time'),
        description = _(u'Time medium format')),

    config.SchemaNode(
        colander.Str(),
        name = 'time_long',
        default = u'%I:%M %p %z',
        title = _(u'Time'),
        description = _(u'Time long format')),

    config.SchemaNode(
        colander.Str(),
        name = 'time_full',
        default = u'%I:%M:%S %p %Z',
        title = _(u'Time'),
        description = _(u'Time full format')),

    title = 'Site formats',
    )


def datetimeFormatter(value, tp='medium', request=None):
    """ datetime format """
    if not isinstance(value, datetime):
        return value

    tz = FORMAT.timezone
    if value.tzinfo is None:
        value = datetime(value.year, value.month, value.day, value.hour,
                         value.minute, value.second, value.microsecond,
                         pytz.utc)

    value = value.astimezone(tz)

    format = '%s %s' % (FORMAT['date_%s' % tp], FORMAT['time_%s' % tp])
    return unicode(value.strftime(str(format)))


def timedeltaFormatter(value, type='short', request=None):
    """ timedelta formatter """
    if not isinstance(value, timedelta):
        return value

    if request is None:
        request = get_current_request()

    if type == 'full':
        hours = value.seconds / 3600
        hs = hours * 3600
        mins = (value.seconds - hs) / 60
        ms = mins * 60
        secs = value.seconds - hs - ms
        frm = []
        translate = get_localizer(request).translate

        if hours:
            frm.append(translate(
                    '${hours} hour(s)', 'ptah.view', {'hours': hours}))
        if mins:
            frm.append(translate(
                    '${mins} min(s)', 'ptah.view', {'mins': mins}))
        if secs:
            frm.append(translate(
                    '${secs} sec(s)', 'ptah.view', {'secs': secs}))

        return ' '.join(frm)

    elif type == 'medium':
        return str(value)
    elif type == 'seconds':
        s = value.seconds + value.microseconds / 1000000.0
        return '%2.4f' % s
    else:
        return str(value).split('.')[0]


_size_types = {
    'b': (1.0, 'B'),
    'k': (1024.0, 'Kb'),
    'm': (1048576.0, 'Mb'),
}


def sizeFormatter(value, type='k', request=None):
    """ size formatter """
    if not isinstance(value, (int, float)):
        return value

    f, t = _size_types.get(type, (1024.0, 'Kb'))

    if t == 'B':
        return '%.0f %s' % (value / f, t)

    return '%.2f %s' % (value / f, t)


format['datetime'] = datetimeFormatter
format['timedelta'] = timedeltaFormatter
format['size'] = sizeFormatter



""" mail settings """
import os.path
from email import Encoders
from email.MIMEText import MIMEText
from email.MIMEMultipart import MIMEMultipart
from email.MIMENonMultipart import MIMENonMultipart
from email.Utils import formatdate, formataddr
from email.Header import make_header

from ptah.settings import MAIL


class MailGenerator(object):
    """ mail generator """

    def __init__(self, context):
        self._headers = {}
        self.context = context

    def _add_header(self, header, value, encode=False):
        self._headers[header] = (header, value, encode)

    def set_headers(self, message):
        charset = str(self.context.charset)

        extra = list(self.context.get_headers())
        for key, val, encode in self._headers.values() + extra:
            if encode:
                message[key] = make_header(((val, charset),))
            else:
                message[key] = val

    def get_message(self):
        """ render a mail template """
        context = self.context

        charset = str(context.charset)
        contentType = context.content_type

        mail_body = context.render()
        maintype, subtype = contentType.split('/')

        message = MIMEText(
            mail_body.encode(charset), subtype, charset)

        return message

    def get_attachments(self):
        attachments = []

        # attach files
        for data, content_type, filename, disposition in \
                self.context.get_attachments():
            maintype, subtype = content_type.split('/')

            msg = MIMENonMultipart(maintype, subtype)

            msg.set_payload(data)
            if filename:
                msg['Content-Id'] = '<%s@ptah>' % filename
                msg['Content-Disposition'] = '%s; filename="%s"' % (
                    disposition, filename)

            Encoders.encode_base64(msg)

            attachments.append(msg)

        return attachments

    def message(self, multipart_format='mixed', *args, **kw):
        context = self.context

        # generate message
        message = self.get_message()

        # generate attachments
        attachments = self.get_attachments()
        if attachments:
            # create multipart message
            root = MIMEMultipart(multipart_format)

            # insert headers
            self.set_headers(root)

            # create message with attachments
            related = MIMEMultipart('related')
            related.attach(message)

            for attach in attachments:
                disposition = attach['Content-Disposition'].split(';')[0]
                if disposition == 'attachment':
                    root.attach(attach)
                else:
                    related.attach(attach)

            root.attach(related)
            message = root

        # alternative
        alternatives = self.context.get_alternative()
        if alternatives:
            mainmessage = MIMEMultipart('alternative')
            mainmessage.attach(message)

            for msg in alternatives:
                mainmessage.attach(MailGenerator(msg).message(
                        multipart_format, *args, **kw))

            message = mainmessage

        # default headers
        self._add_header('Subject', context.subject, True)

        self.set_headers(message)
        return message

    def __call__(self, multipart_format='mixed', *args, **kw):
        context = self.context
        message = self.message(multipart_format, *args, **kw)

        message['Date'] = formatdate()
        message['Message-ID'] = context.message_id

        if not message.get('To') and context.to_address:
            message['To'] = context.to_address

        if not message.get('From') and context.from_address:
            message['From'] = formataddr(
                (context.from_name, context.from_address))

        return message


class MailTemplate(object):
    """ mail template with base features """

    subject = u''
    charset = u'utf-8'
    content_type = u'text/plain'
    message_id = None
    template = None

    from_name = ''
    from_address = ''
    to_address = ''
    return_address = ''
    errors_address = ''

    def __init__(self, context, request, **kwargs):
        self.__dict__.update(kwargs)

        self.context = context
        self.request = request

        self._files = []
        self._headers = {}
        self._alternative = []

    def add_header(self, header, value, encode=False):
        self._headers[header] = (header, value, encode)

    def has_header(self, header):
        header = header.lower()
        for key in self._headers.keys():
            if key.lower() == header:
                return True

        return False

    def get_headers(self):
        return self._headers.values()

    def add_attachment(self, file_data, content_type,
                       filename, disposition='attachment'):
        self._files.append((file_data, content_type,
                            wrap_filename(filename), disposition))

    def get_attachments(self):
        return self._files

    def add_alternative(self, template):
        self._alternative.append(template)

    def get_alternative(self):
        return self._alternative

    def update(self):
        if not self.from_name:
            self.from_name = MAIL.from_name
        if not self.from_address:
            self.from_address = MAIL.from_address

    def render(self):
        kwargs = {'view': self,
                  'context': self.context,
                  'request': self.request}

        return self.template(**kwargs)

    def send(self, email=None, **kw):
        if email:
            self.to_address = email

        message = self(**kw)

        MAIL.Mailer.send(message['from'], message['to'], message)

    def __call__(self, **kw):
        for key, value in kw.items():
            if type(value) is tuple:
                self.add_header(key, value[0], value[1])
            else:
                self.add_header(key, value)

        self.update()
        return MailGenerator(self)()


def wrap_filename(f_name):
    dir, f_name = os.path.split(f_name)
    f_name = f_name.split('\\')[-1].split('/')[-1]
    for key in '~,\'':
        f_name = f_name.replace(key, '_')

    return f_name



""" ptah.view interfaces """
from zope import interface
from pyramid.interfaces import IView as IPyramidView


class IPtahView(IPyramidView):
    """ ptah view """

    __name__ = interface.Attribute('Name')
    __parent__ = interface.Attribute('Parent')

    context = interface.Attribute('Context')
    request = interface.Attribute('Request')

    def update():
        """Update the snippet data."""

    def __call__():
        """ render view """


class ISnippet(interface.Interface):
    """ snippet """

    context = interface.Attribute('Context')

    request = interface.Attribute('Request')

    template = interface.Attribute('Template')

    def update():
        """Update the snippet data."""

    def render():
        """Render the snippet."""

    def __call__():
        """Update and render snippet"""


class ILayout(interface.Interface):
    """ layout """

    template = interface.Attribute('Layout template')

    view = interface.Attribute('Initial view')
    viewcontext = interface.Attribute('Initial view context')

    template = interface.Attribute('Layout template')

    def update():
        """Update the layout data """

    def render(content, **kwargs):
        """Render the layout """

    def __call__(content, layout=None, view=None, *args, **kw):
        """ build layout tree and render """


# status message
class IMessage(interface.Interface):
    """ message """

    def render(message):
        """ render message """


class IStatusMessage(interface.Interface):
    """ message service """

    def add(text, type='info'):
        """ add message text as message to service """

    def clear():
        """ return all mesasges and clear """

    def messages():
        """ return all messages """



from ptah import config
from base import Base


class TestAction(Base):

    def tearDown(self):
        config.cleanup_system(self.__class__.__module__)
        super(TestAction, self).tearDown()

    def test_cms_action_reg(self):
        import ptah.cms

        class Test(object):
            @ptah.cms.action
            def update(self, **data): # pragma: no cover
                pass

        self.assertTrue(hasattr(Test, '__ptahcms_actions__'))

        actions = Test.__ptahcms_actions__
        self.assertIn('update', actions)
        self.assertEqual(actions['update'], ('update', ptah.cms.View))

    def test_cms_action_reg_name(self):
        import ptah.cms

        class Test(object):
            @ptah.cms.action(name='test')
            def update(self, **data): # pragma: no cover
                pass

        actions = Test.__ptahcms_actions__
        self.assertIn('test', actions)
        self.assertEqual(actions['test'], ('update', ptah.cms.View))

    def test_cms_action_reg_permission(self):
        import ptah.cms

        class Test(object):
            @ptah.cms.action(permission='perm')
            def update(self, **data): # pragma: no cover
                pass

        actions = Test.__ptahcms_actions__
        self.assertEqual(actions['update'], ('update', 'perm'))

    def test_cms_action_inherit(self):
        import ptah.cms
        from ptah.cms.security import build_class_actions

        class Test(object):
            @ptah.cms.action(permission='perm')
            def update(self, **data): # pragma: no cover
                pass

        class Test2(Test):
            pass

        build_class_actions(Test2)

        actions = Test2.__ptahcms_actions__

        self.assertEqual(actions['update'], ('update', 'perm'))
        self.assertIsNot(actions, Test.__ptahcms_actions__)

    def test_cms_action_inherit2(self):
        import ptah.cms
        from ptah.cms.security import build_class_actions

        class Test(object):
            @ptah.cms.action
            def update(self, **data): # pragma: no cover
                pass

        class Test2(Test):
            @ptah.cms.action
            def create(self, **data): # pragma: no cover
                pass

        build_class_actions(Test2)

        actions = Test2.__ptahcms_actions__

        self.assertEqual(len(actions), 2)
        self.assertIn('update', actions)
        self.assertIn('create', actions)


class TestWrapper(Base):

    def tearDown(self):
        config.cleanup_system(self.__class__.__module__)
        super(TestWrapper, self).tearDown()

    def test_cms_wrapper_not_found(self):
        import ptah.cms
        from ptah.cms.security import NodeWrapper

        class Test(object):
            @ptah.cms.action
            def update(self, **data): # pragma: no cover
                pass

        wrapper = NodeWrapper(Test())
        self.assertRaises(
            ptah.cms.NotFound, wrapper.__getattr__, 'unknown')

    def test_cms_wrapper_forbidden(self):
        import ptah, ptah.cms
        from ptah.cms.security import NodeWrapper

        class Test(object):
            @ptah.cms.action(permission=ptah.NOT_ALLOWED)
            def update(self, *args, **data): # pragma: no cover
                return 'test'

        wrapper = NodeWrapper(Test())

        self.assertRaises(
            ptah.cms.Forbidden, wrapper.__getattr__, 'update')

    def test_cms_wrapper(self):
        import ptah, ptah.cms
        from ptah.cms.security import NodeWrapper

        class Test(object):
            @ptah.cms.action(permission=ptah.NO_PERMISSION_REQUIRED)
            def update(self, *args, **data):
                return 'test'

        wrapper = NodeWrapper(Test())
        self.assertEqual(wrapper.update(), 'test')


class TestCms(Base):

    def _setup_ptah(self):
        pass

    def tearDown(self):
        config.cleanup_system(self.__class__.__module__)
        super(TestCms, self).tearDown()

    def test_cms_not_found(self):
        import ptah.cms

        self.assertRaises(ptah.cms.NotFound, ptah.cms.wrap, 'unknown')
        self.assertRaises(ptah.cms.NotFound, ptah.cms.wrap, None)

    def test_cms(self):
        import ptah
        from ptah.cms.security import wrap, NodeWrapper

        class Test(ptah.cms.Content):
            __uri_factory__ = ptah.UriFactory('test')

            @ptah.cms.action(permission=ptah.NO_PERMISSION_REQUIRED)
            def update(self, *args, **data): # pragma: no cover
                pass

        t = Test()
        wrapper = wrap(t)
        self.assertIsInstance(wrapper, NodeWrapper)
        self.assertIs(wrapper._content, t)

    def test_cms_2(self):
        import ptah
        from ptah.cms.security import wrap, NodeWrapper

        class Test(ptah.cms.Content):
            __uri_factory__ = ptah.UriFactory('test')

            @ptah.cms.action(permission=ptah.NO_PERMISSION_REQUIRED)
            def update(self, *args, **data): # pragma: no cover
                pass

        t = Test()

        @ptah.resolver('test')
        def res(uri):
            return t

        self._init_ptah()

        wrapper = wrap('test:1')
        self.assertIsInstance(wrapper, NodeWrapper)
        self.assertIs(wrapper._content, t)



""" directives tests """
import sys
import unittest

from pyramid import testing
from zope import interface
from zope.interface.registry import Components
from zope.interface.interfaces import IObjectEvent

import ptah
from ptah import config
from ptah.config import directives
from ptah.config.api import objectEventNotify


class BaseTesting(unittest.TestCase):

    def _init_ptah(self, settings={}, pconfig=None, *args, **kw):
        if pconfig is None:
            pconfig = self.config
        ptah.config.initialize(pconfig, ('ptah.config', self.__class__.__module__))

    def setUp(self):
        self.config = testing.setUp()
        self.registry = self.config.registry

    def tearDown(self):
        config.cleanup_system(self.__class__.__module__)

        global TestClass, testAdapter, testHandler
        try:
            del TestClass
        except:
            pass
        try:
            del testAdapter
        except:
            pass
        try:
            del testHandler
        except:
            pass

        testing.tearDown()


class IContext(interface.Interface):
    pass


class IContext2(interface.Interface):
    pass


class IAdapter(interface.Interface):
    pass


class Context(object):

    def __init__(self, iface):
        interface.directlyProvides(self, iface)


class TestAdaptsDirective(BaseTesting):

    def test_adapts_err(self):
        global TestClass

        class TestClass(object):
            config.adapter(IContext)
            config.adapter(IContext)
            interface.implements(IAdapter)

        try:
            self._init_ptah()
        except Exception, e:
            pass

        s = str(e)
        self.assertTrue(isinstance(e, config.ConflictError))
        self.assertTrue('ptah.config:adapter' in s)
        self.assertTrue('test_directives.py' in s)

    def test_adapts(self):
        global TestClass

        class TestClass(object):
            config.adapter(IContext)
            interface.implements(IAdapter)

            def __init__(self, context):
                pass

        self._init_ptah()

        sm = self.registry
        adapters = sm.adapters.lookupAll((IContext,), IAdapter)

        self.assertTrue(len(adapters) == 1)
        self.assertTrue(adapters[0][0] == '')
        self.assertTrue(adapters[0][1] is TestClass)

        adapter = IAdapter(Context(IContext))
        self.assertTrue(isinstance(adapter, TestClass))

    def test_adapts_named(self):
        global TestClass

        class TestClass(object):
            config.adapter(IContext, name='test')
            interface.implements(IAdapter)

        self._init_ptah()

        sm = self.registry
        adapters = sm.adapters.lookupAll((IContext,), IAdapter)

        self.assertTrue(adapters[0][0] == 'test')
        self.assertTrue(adapters[0][1] is TestClass)

    def test_adapts_multiple(self):
        global TestClass

        class TestClass(object):
            config.adapter(IContext, IContext2)
            interface.implements(IAdapter)

        self._init_ptah()

        sm = self.registry
        adapters = sm.adapters.lookupAll((IContext, IContext2), IAdapter)

        self.assertTrue(adapters[0][0] == '')
        self.assertTrue(adapters[0][1] is TestClass)

    def test_adapts_multiple_named(self):
        global TestClass

        class TestClass(object):
            config.adapter(IContext, IContext2, name='test')
            interface.implements(IAdapter)

        self._init_ptah()

        sm = self.registry
        adapters = sm.adapters.lookupAll((IContext, IContext2), IAdapter)

        self.assertTrue(adapters[0][0] == 'test')
        self.assertTrue(adapters[0][1] is TestClass)

    def test_adapts_reinitialize(self):
        global TestClass

        class TestClass(object):
            config.adapter(IContext)
            interface.implements(IAdapter)

        self._init_ptah()

        # reinstall
        config.cleanup_system()

        sm = self.registry
        sm.__init__('base')
        adapters = sm.adapters.lookupAll((IContext,), IAdapter)
        self.assertTrue(len(adapters) == 0)

        self._init_ptah()

        sm = self.registry
        adapters = sm.adapters.lookupAll((IContext,), IAdapter)

        self.assertTrue(len(adapters) == 1)
        self.assertTrue(adapters[0][1] is TestClass)


class TestAdapterDirective(BaseTesting):

    def test_adapter(self):
        global testAdapter

        @config.adapter(IContext)
        @config.adapter(IContext2)
        @interface.implementer(IAdapter)
        def testAdapter(context):  # pragma: no cover
            pass

        self._init_ptah()

        sm = self.registry
        adapters = sm.adapters.lookupAll((IContext,), IAdapter)

        self.assertTrue(len(adapters) == 1)
        self.assertTrue(adapters[0][1] is testAdapter)

        adapters = sm.adapters.lookupAll((IContext2,), IAdapter)

        self.assertTrue(len(adapters) == 1)
        self.assertTrue(adapters[0][1] is testAdapter)

    def test_adapter_named(self):
        global testAdapter

        @config.adapter(IContext, name='test')
        @interface.implementer(IAdapter)
        def testAdapter(context):  # pragma: no cover
            pass

        self._init_ptah()

        sm = self.registry
        adapters = sm.adapters.lookupAll((IContext,), IAdapter)

        self.assertTrue(len(adapters) == 1)
        self.assertTrue(adapters[0][0] == 'test')
        self.assertTrue(adapters[0][1] is testAdapter)

    def test_adapter_multple(self):
        global testAdapter

        @config.adapter(IContext, IContext2, name='test')
        @interface.implementer(IAdapter)
        def testAdapter(context):  # pragma: no cover
            pass

        self._init_ptah()

        sm = self.registry
        adapters = sm.adapters.lookupAll((IContext, IContext2), IAdapter)

        self.assertTrue(len(adapters) == 1)
        self.assertTrue(adapters[0][0] == 'test')
        self.assertTrue(adapters[0][1] is testAdapter)

    def test_adapter_errs1(self):
        global testAdapter

        try:
            @config.adapter(IContext, name='test')
            def testAdapter(context):  # pragma: no cover
                pass
        except Exception, e:
            pass

        self.assertTrue(isinstance(e, TypeError))

    def test_adapter_errs2(self):
        global testAdapter

        @config.adapter(IContext)
        @config.adapter(IContext)
        @interface.implementer(IAdapter)
        def testAdapter(context):  # pragma: no cover
            pass

        self.assertRaises(config.ConflictError, self._init_ptah)

    def test_adapter_reinitialize(self):
        global testAdapter

        @config.adapter(IContext)
        @config.adapter(IContext2)
        @interface.implementer(IAdapter)
        def testAdapter(context):  # pragma: no cover
            pass

        self._init_ptah()

        # reinstall
        config.cleanup_system()

        sm = self.registry
        sm.__init__('base')

        self._init_ptah()

        adapters = sm.adapters.lookupAll((IContext,), IAdapter)

        self.assertTrue(len(adapters) == 1)
        self.assertTrue(adapters[0][1] is testAdapter)

        adapters = sm.adapters.lookupAll((IContext2,), IAdapter)

        self.assertTrue(len(adapters) == 1)
        self.assertTrue(adapters[0][1] is testAdapter)


class TestSubscriberDirective(BaseTesting):

    def test_subscriber(self):
        global testSubscriber

        events = []

        @config.subscriber(IContext)
        def testHandler(*args):
            events.append(args)

        self._init_ptah()

        sm = self.registry
        sm.subscribers((Context(IContext),), None)

        self.assertTrue(len(events) == 1)

        sm.subscribers((Context(IContext2),), None)
        self.assertTrue(len(events) == 1)

    def test_subscriber_multple(self):
        global testSubscriber

        events = []

        @config.subscriber(IContext)
        @config.subscriber(IContext2)
        def testHandler(*args):
            events.append(args)

        self._init_ptah()

        sm = self.registry
        sm.subscribers((Context(IContext),), None)
        sm.subscribers((Context(IContext2),), None)

        self.assertTrue(len(events) == 2)

    def test_subscriber_object(self):
        global testSubscriber
        from zope.interface.interfaces import IObjectEvent, ObjectEvent

        events = []

        @config.subscriber(IContext, IObjectEvent)
        def testSubscriber(*args):
            events.append(args)

        self._init_ptah()

        sm = self.config.registry
        sm.subscribers((ObjectEvent(Context(IContext)),), None)

        self.assertTrue(len(events) == 1)
        self.assertTrue(len(events[0]) == 2)

    def test_subscriber_reinitialize(self):
        global testSubscriber

        events = []

        @config.subscriber(IContext)
        def testSubscriber(*args):
            events.append(args)

        self._init_ptah()

        # reinstall
        config.cleanup_system()

        sm = self.registry
        sm.__init__('base')

        self._init_ptah()

        sm = self.registry
        sm.subscribers((Context(IContext),), None)
        self.assertTrue(len(events) == 1)


class TestExtraDirective(BaseTesting):

    def test_scan_unknown(self):
        self.assertRaises(ImportError,  directives.scan, 'unknown', [])

    def test_directive_info_limit_scope(self):
        self.assertRaises(
            TypeError,
            directives.DirectiveInfo, 2, allowed_scope=('class',))

    def test_directive_info_context(self):
        info = directives.DirectiveInfo(0)
        info.scope = 'module'

        self.assertEqual(info.module,
                         sys.modules[self.__class__.__module__])
        self.assertEqual(info.context,
                         sys.modules[self.__class__.__module__])

    def test_api_loadpackage(self):
        from ptah.config import api

        seen = set()
        packages = api.load_package('ptah.config', seen)
        self.assertEqual(packages, ['ptah.config'])

        packages = api.load_package('ptah.config', seen)
        self.assertEqual(packages, [])

    def test_action(self):
        info = config.DirectiveInfo()

        action = config.Action(None, discriminator=('test', ))

        self.assertIsNone(action.hash)

        info.attach(action)

        self.assertIsNotNone(action.hash)
        self.assertIsNotNone(hash(action))
        self.assertRaises(TypeError, info.attach, action)



import transaction
from ptah import config
from datetime import timedelta

from base import Base


class TestTokenType(Base):

    def tearDown(self):
        config.cleanup_system(self.__class__.__module__)
        super(TestTokenType, self).tearDown()

    def test_token(self):
        from ptah import token

        tt = token.TokenType('unique-id', timedelta(minutes=20))
        self._init_ptah()

        t = token.service.generate(tt, 'data')
        transaction.commit()

        self.assertEqual(token.service.get(t), 'data')
        self.assertEqual(token.service.get_bydata(tt, 'data'), t)

        token.service.remove(t)
        self.assertEqual(token.service.get(t), None)

    def test_token_type(self):
        from ptah import token

        token.TokenType('unique-id', timedelta(minutes=20))
        token.TokenType('unique-id', timedelta(minutes=20))

        self.assertRaises(config.ConflictError, self._init_ptah)

    def test_token_remove_expired(self):
        pass



""" blob storage implementation """
import os
import sqlalchemy as sqla
from zope import interface

import ptah
from node import Node, Session
from interfaces import IBlob, IBlobStorage


class Blob(Node):
    interface.implements(IBlob)

    __tablename__ = 'ptah_blobs'
    __uri_factory__ = ptah.UriFactory('blob-sql')

    __id__ = sqla.Column('id', sqla.Integer,
                         sqla.ForeignKey('ptah_nodes.id'), primary_key=True)

    mimetype = sqla.Column(sqla.String(), default='')
    filename = sqla.Column(sqla.String(), default='')
    size = sqla.Column(sqla.Integer, default=0)
    data = sqla.orm.deferred(sqla.Column(sqla.LargeBinary))

    def read(self):
        return self.data

    def write(self, data):
        self.data = data
        self.size = len(data)

    def updateMetadata(self, mimetype=None, filename=None, **md):
        if mimetype is not None:
            self.mimetype = mimetype

        if filename is not None:
            self.filename = filename

    def info(self):
        info = super(Blob, self).info()

        info['size'] = self.size
        info['mimetype'] = self.mimetype
        info['filename'] = self.filename
        return info


class BlobStorage(object):
    interface.implements(IBlobStorage)

    _sql_get = ptah.QueryFreezer(
        lambda: Session.query(Blob)
            .filter(Blob.__uri__ == sqla.sql.bindparam('uri')))

    _sql_get_by_parent = ptah.QueryFreezer(
        lambda: Session.query(Blob)
            .filter(Blob.__parent_uri__ == sqla.sql.bindparam('parent')))

    def create(self, parent=None):
        blob = Blob(__parent__=parent)
        Session.add(blob)
        Session.flush()

        return blob

    def add(self, data, parent=None, **metadata):
        blob = self.create(parent)

        data.seek(0)

        blob.data = data.read()
        blob.updateMetadata(**metadata)

        data.seek(0, os.SEEK_END)
        blob.size = data.tell()

        return blob

    def get(self, uri):
        """SQL Blob resolver"""
        return self._sql_get.first(uri=uri)

    def getByParent(self, parent):
        return self._sql_get_by_parent.first(parent=parent)

    def replace(self, uri, data, **metadata): # pragma: no cover
        pass

    def remove(self, uri): # pragma: no cover
        pass


blobStorage = BlobStorage()

ptah.register_uri_resolver('blob-sql', blobStorage.get)



import transaction
from base import Base


class TestNode(Base):

    def test_node_ctor(self):
        import ptah.cms

        # it not possible to instatiate Node
        self.assertRaises(TypeError, ptah.cms.Node)

    def test_node(self):
        import ptah.cms

        class MyContent(ptah.cms.Node):
            __uri_factory__ = ptah.UriFactory('test')

        content = MyContent()
        _uri = content.__uri__
        ptah.cms.Session.add(content)
        transaction.commit()

        c = ptah.cms.Session.query(ptah.cms.Node).filter(
            ptah.cms.Node.__uri__ == _uri).one()

        self.assertTrue(isinstance(c, ptah.cms.Node))

    def test_polymorphic_node(self):
        import ptah.cms

        class MyContent(ptah.cms.Node):
            __mapper_args__ = {'polymorphic_identity': 'mycontent'}
            __uri_factory__ = ptah.UriFactory('test')

        content = MyContent()
        _uri = content.__uri__
        ptah.cms.Session.add(content)
        transaction.commit()

        c = ptah.cms.Session.query(ptah.cms.Node).filter(
            ptah.cms.Node.__uri__ == _uri).one()

        self.assertTrue(isinstance(c, MyContent))

    def test_node_parent(self):
        import ptah.cms

        class MyContent(ptah.cms.Node):
            __mapper_args__ = {'polymorphic_identity': 'mycontent'}
            __uri_factory__ = ptah.UriFactory('test')

        parent = MyContent()
        parent_uri = parent.__uri__

        content = MyContent(__parent__ = parent)
        __uri = content.__uri__

        ptah.cms.Session.add(parent)
        ptah.cms.Session.add(content)
        transaction.commit()

        c = ptah.cms.Session.query(ptah.cms.Node).filter(
            ptah.cms.Node.__uri__ == __uri).one()

        self.assertTrue(c.__parent_uri__ == parent_uri)
        self.assertTrue(c.__parent_ref__.__uri__ == parent_uri)

    def test_node_local_roles(self):
        import ptah

        class MyContent(ptah.cms.Node):
            __uri_factory__ = ptah.UriFactory('test')
            __mapper_args__ = {'polymorphic_identity': 'mycontent'}

        content = MyContent()
        __uri = content.__uri__

        self.assertTrue(ptah.ILocalRolesAware.providedBy(content))

        ptah.cms.Session.add(content)
        transaction.commit()

        c = ptah.cms.Session.query(ptah.cms.Node).filter(
            ptah.cms.Node.__uri__ == __uri).one()

        c.__local_roles__['userid'] = ('role:1',)
        transaction.commit()

        c = ptah.cms.Session.query(ptah.cms.Node).filter(
            ptah.cms.Node.__uri__ == __uri).one()
        self.assertTrue(c.__local_roles__ == {u'userid': [u'role:1']})

    def test_node_owners(self):
        import ptah

        class MyContent(ptah.cms.Node):
            __mapper_args__ = {'polymorphic_identity': 'mycontent'}
            __uri_factory__ = ptah.UriFactory('test')

        content = MyContent()
        __uri = content.__uri__

        self.assertTrue(ptah.IOwnersAware.providedBy(content))

        ptah.cms.Session.add(content)
        transaction.commit()

        c = ptah.cms.Session.query(ptah.cms.Node).filter(
            ptah.cms.Node.__uri__ == __uri).one()

        c.__owner__ = 'userid'
        transaction.commit()

        c = ptah.cms.Session.query(ptah.cms.Node).filter(
            ptah.cms.Node.__uri__ == __uri).one()
        self.assertTrue(c.__owner__ == u'userid')

    def test_node_permissions(self):
        import ptah

        class MyContent(ptah.cms.Node):
            __mapper_args__ = {'polymorphic_identity': 'mycontent'}
            __uri_factory__ = ptah.UriFactory('test')

        content = MyContent()
        __uri = content.__uri__

        self.assertTrue(ptah.IACLsAware.providedBy(content))

        ptah.cms.Session.add(content)
        transaction.commit()

        c = ptah.cms.Session.query(ptah.cms.Node).filter(
            ptah.cms.Node.__uri__ == __uri).one()

        c.__acls__.append('map1')
        transaction.commit()

        c = ptah.cms.Session.query(ptah.cms.Node).filter(
            ptah.cms.Node.__uri__ == __uri).one()
        self.assertTrue(c.__acls__ == [u'map1'])



""" This is an example of useing form (imperative style). """
from paste.httpserver import serve

import ptah
from ptah import cms
from ptah.cms import restaction, View, ModifyContent


@restaction('extra-info', ptah_cms.Content, permission=View)
def extraInfo(content, request):
    """ __doc__ is used for action description """

    return {'title': content.title,
            'email': 'ptah@ptahproject.org',
            'message': 'Ptah rest api'}


@restaction('protected-info', ptah_cms.Content, permission=ModifyContent)
def protectedInfo(content, request):
    """ protected rest action """

    return {'title': content.title,
            'email': 'ptah@ptahproject.org',
            'message': 'Ptah rest api'}


if __name__ == '__main__':
    """ ...

    """
    app = ptah.make_wsgi_app({'settings':r'./ptah.ini'})
    serve(app, host='0.0.0.0')



""" paste commands """
import argparse
import ConfigParser
import StringIO
import textwrap

from collections import OrderedDict
from pyramid.config import Configurator

from ptah import config
from ptah.config.settings import get_settings


grpTitleWrap = textwrap.TextWrapper(
    initial_indent='* ',
    subsequent_indent='  ')

grpDescriptionWrap = textwrap.TextWrapper(
    initial_indent='    ',
    subsequent_indent='    ')

nameWrap = textwrap.TextWrapper(
    initial_indent='  - ',
    subsequent_indent='    ')

nameTitleWrap = textwrap.TextWrapper(
    initial_indent='       ',
    subsequent_indent='       ')

nameDescriptionWrap = textwrap.TextWrapper(
    initial_indent=' * ',
    subsequent_indent='')


def settingsCommand(init=True):
    if init:  # pragma: no cover
        config.initialize(Configurator(), autoinclude=True)

    args = SettingsCommand.parser.parse_args()
    cmd = SettingsCommand(args)
    cmd.run()


class SettingsCommand(object):
    """ 'settings' paste command"""

    parser = argparse.ArgumentParser(description="ptah settings management")
    parser.add_argument('-a', '--all', action="store_true",
                        dest='all',
                        help='List all registered settings')
    parser.add_argument('-l', '--list',
                        dest='section', default='',
                        help='List registered settings')
    parser.add_argument('-p', '--print', action="store_true",
                        dest='printcfg',
                        help='Print default settings in ConfigParser format')

    def __init__(self, args):
        self.options = args

    def run(self):
        # print defaults
        if self.options.printcfg:
            data = get_settings().export(True)

            parser = ConfigParser.ConfigParser(dict_type=OrderedDict)
            items = data.items()
            items.sort()
            for key, val in items:
                parser.set(ConfigParser.DEFAULTSECT, key, val)

            fp = StringIO.StringIO()
            try:
                parser.write(fp)
            finally:
                pass

            print fp.getvalue()
            return

        if self.options.all:
            section = ''
        else:
            section = self.options.section

        # print description
        groups = get_settings().items()
        groups.sort()

        for name, group in groups:
            if section and name != section:
                continue

            print ''
            title = group.title or name

            print grpTitleWrap.fill(title)
            if group.description:
                print grpDescriptionWrap.fill(group.description)

            print
            for node in group.schema:
                default = '<required>' if node.required else node.default
                print nameWrap.fill(
                    '%s.%s: %s (%s: %s)' % (
                        name, node.name, node.title,
                        node.typ.__class__.__name__, default))

                print nameTitleWrap.fill(node.description)
                print



""" Form buttons """
import re
import sys

from ptah import view
from collections import OrderedDict

AC_DEFAULT = 0
AC_PRIMARY = 1
AC_DANGER = 2
AC_SUCCESS = 3
AC_INFO = 4

css = {
    AC_PRIMARY: 'primary',
    AC_DANGER: 'danger',
    AC_SUCCESS: 'success',
    AC_INFO: 'info'}


class Button(object):
    """A simple button in a form."""

    lang = None
    readonly = False
    alt = None
    accesskey = None
    disabled = False
    tabindex = None
    klass = u'btn'

    template = view.template("ptah.form:templates/submit.pt")

    def __init__(self, name='submit', title=None, action=None, actionName=None,
                 actype=AC_DEFAULT, condition=None, **kw):
        self.__dict__.update(kw)

        if title is None:
            title = name.capitalize()
        name = re.sub(r'\s', '_', name)

        self.name = name
        self.title = title
        self.action = action
        self.actionName = actionName
        self.actype = actype
        self.condition = condition

    def __repr__(self):
        return '<%s %r : %r>' % (self.__class__.__name__, self.name,
                                 self.title)

    def __call__(self, context):
        if self.actionName is not None:
            return getattr(context, self.actionName)()
        elif self.action is not None:
            return self.action(context)
        else:
            raise TypeError("Action is not specified")

    def bind(self, prefix, params, context, request):
        widget = self.__class__.__new__(self.__class__)
        widget.__dict__.update(self.__dict__)

        widget.id = str(prefix + widget.name).replace('.', '-')
        widget.name = str(prefix + widget.name)
        widget.params = params
        widget.context = context
        widget.request = request
        widget.klass = u'%s %s' % (widget.klass, css.get(widget.actype, ''))
        return widget

    def activated(self):
        return self.params.get(self.name, None) is not None

    def render(self):
        return self.template(context=self, request=self.request)


class Buttons(OrderedDict):
    """Buttons manager"""

    def __init__(self, *args):
        super(Buttons, self).__init__()

        buttons = []
        for arg in args:
            if isinstance(arg, Buttons):
                buttons += arg.values()
            else:
                buttons.append(arg)

        self.add(*buttons)

    def add(self, *btns):
        for btn in btns:
            if btn.name in self:
                raise ValueError("Duplicate name", btn.name)

            self[btn.name] = btn

    def add_action(self, title, **kwargs):
        # Add the title to button constructor keyword arguments
        kwargs['title'] = title
        if 'name' not in kwargs:
            kwargs['name'] = create_btn_id(title)

        button = Button(**kwargs)

        self.add(button)

        return button

    def __add__(self, other):
        return self.__class__(self, other)


class Actions(OrderedDict):
    """Form actions manager"""

    prefix = 'buttons.'

    def __init__(self, form, request):
        self.form = form
        self.request = request

        super(Actions, self).__init__()

    def update(self):
        form = self.form
        params = form.form_params()

        # Create a unique prefix.
        prefix = '%s%s' % (form.prefix, self.prefix)

        # Walk through each node, making a widget out of it.
        for field in self.form.buttons.values():
            if field.condition and not field.condition(form):
                continue

            self[field.name] = field.bind(prefix, params, form, self.request)

    def execute(self):
        result = None
        executed = False
        for action in self.values():
            if action.activated():
                executed = True
                result = action(self.form)

        if executed:
            self.clear()
            self.update()

        return result


_identifier = re.compile('[A-Za-z][a-zA-Z0-9_]*$')


def create_btn_id(name):
    if _identifier.match(name):
        return str(name).lower()
    return name.encode('utf-8').encode('hex')


def button(title, **kwargs):
    # install buttons manager
    f_locals = sys._getframe(1).f_locals

    buttons = f_locals.get('buttons')
    if buttons is None:
        buttons = Buttons()
        f_locals['buttons'] = buttons

    btn = buttons.add_action(title, **kwargs)

    def createHandler(func):
        btn.actionName = func.__name__
        return func

    return createHandler



import inspect
import simplejson as json
from collections import OrderedDict
from ptah import view
from ptah.view.view import PyramidView

from zope.interface import providedBy
from pyramid.interfaces import IView, IViewClassifier
from pyramid.httpexceptions import HTTPNotFound


class LayoutPreview(view.View):
    view.pview('layout-preview.html', layout=None)

    view = None
    layout = None
    colors = ['green', 'yellow', 'blue', 'gray', 'black', 'black']

    def update(self):
        request = self.request
        context = self.context

        self.view_name = view_name = request.GET.get('view', '')

        adapters = request.registry.adapters

        view_callable = adapters.lookup(
            (IViewClassifier, providedBy(request), providedBy(context)),
            IView, name=view_name, default=None)

        if not isinstance(view_callable, PyramidView):
            raise HTTPNotFound()

        view_renderer = None
        layout_renderer = None

        for r in view_callable.renderers:
            if isinstance(r, view.LayoutRenderer):
                layout_renderer = r
            if isinstance(r, view.ViewRenderer):
                view_renderer = r

        if layout_renderer is not None:
            self.layout = view.query_layout_chain(
                context, request, layout_renderer.layout)

        self.view = view_renderer
        self.action = view_callable.__config_action__
        self.info = view_callable.__config_action__.info

        if self.layout is None:
            raise HTTPNotFound()

    def build_layout_info(self, layout):
        action = layout.__config_action__
        info = action.info

        data = OrderedDict(
            (('name', layout.name),
             ('parent-layout', layout.layout),
             ('class', '%s.%s'%(layout.__class__.__module__,
                                layout.__class__.__name__)),
             ('python-module', info.module.__name__),
             ('python-module-location', info.module.__file__),
             ('python-module-line', info.codeinfo[1]),
             ('template', 'unset'),
             ('context', '%s.%s'%(layout.context.__class__.__module__,
                                  layout.context.__class__.__name__)),
             ('context-path', self.request.resource_url(layout.context)),
             ))

        if layout.template:
            data['template'] = str(layout.template)

        return data

    def build_view_info(self, view):
        info = self.info
        action = self.action

        if inspect.isclass(info.context):
            factory = info.context
        else:
            factory = action.args[0]

        if inspect.isclass(factory):
            factory = '%s.%s'%(factory.__module__,
                               factory.__name__)
        else:
            factory = factory.__name__

        data = OrderedDict(
            (('name', self.view_name),
             ('view-factory', factory),
             ('python-module', info.module.__name__),
             ('python-module-location', info.module.__file__),
             ('python-module-line', info.codeinfo[1]),
             ('template', 'unset'),
             ('context', '%s.%s'%(self.context.__class__.__module__,
                                  self.context.__class__.__name__)),
             ('context-path', self.request.resource_url(self.context)),
             ))

        if hasattr(view, 'template'):
            data['template'] = str(view.template)

        return data

    def render(self):
        res = []

        content = self.view(self.context, self.request, '')
        content = u'\n<!-- view:\n%s \n-->\n'\
                  u'<div style="border: 2px solid red">%s</div>'%(
            json.dumps(self.build_view_info(self.view), indent=True),
            content)

        for layout in self.layout:
            layout.update()

            res = layout.render(content)
            idx = self.layout.index(layout)
            content = u'\n<!-- layout:\n%s \n-->\n'\
                      u'<div style="border: 4px solid %s">%s</div>'%(
                json.dumps(self.build_layout_info(layout), indent=True),
                self.colors[idx], res)

        return content



import time
import ptah
import transaction
from ptah import config
from datetime import datetime

from base import Base


class TestContent(Base):

    def tearDown(self):
        config.cleanup_system(self.__class__.__module__)
        super(TestContent, self).tearDown()

    def _make_app(self):
        global ApplicationRoot
        class ApplicationRoot(ptah.cms.ApplicationRoot):
            __type__ = ptah.cms.Type('app')

        ApplicationRoot.__type__.cls = ApplicationRoot

        return ApplicationRoot

    def test_content_path(self):
        import ptah.cms
        self._setRequest(self._makeRequest())

        class MyContent(ptah.cms.Content):

            __mapper_args__ = {'polymorphic_identity': 'mycontent'}
            __uri_factory__ = ptah.UriFactory('mycontent')

        ApplicationRoot = self._make_app()

        factory = ptah.cms.ApplicationFactory(
            ApplicationRoot, '/app1', 'root', 'Root App')

        root = factory(self.request)

        content = MyContent(__name__='test',
                            __parent__ = root,
                            __path__ = '%stest/'%root.__path__)
        c_uri = content.__uri__
        ptah.cms.Session.add(content)

        self.assertTrue(
            content.__name__ == 'test')

        self.assertTrue(
            content.__resource_url__(self.request, {}) == '/app1/test/')
        transaction.commit()

        # same content inside same root but in different app factory

        factory2 = ptah.cms.ApplicationFactory(
            ApplicationRoot, '/app2', 'root', 'Root App')
        root = factory2(self.request)

        c = ptah.cms.Session.query(MyContent).filter(
            MyContent.__uri__ == c_uri).one()

        self.assertTrue(
            c.__resource_url__(self.request, {}) == '/app2/test/')

    def test_content_events(self):
        import ptah.cms

        class MyContent(ptah.cms.Content):
            __mapper_args__ = {'polymorphic_identity': 'mycontent'}
            __uri_factory__ = ptah.UriFactory('mycontent')

        content = MyContent()

        config.notify(ptah.cms.ContentCreatedEvent(content))

        self.assertTrue(isinstance(content.created, datetime))
        self.assertTrue(isinstance(content.modified, datetime))
        time.sleep(0.1)

        config.notify(ptah.cms.ContentModifiedEvent(content))
        self.assertTrue(content.modified != content.created)

    def test_content_set_owner_on_create(self):
        import ptah, ptah.cms

        class MyContent(ptah.cms.Content):
            __mapper_args__ = {'polymorphic_identity': 'mycontent'}
            __uri_factory__ = ptah.UriFactory('mycontent')

        content = MyContent()

        config.notify(ptah.cms.ContentCreatedEvent(content))

        self.assertEqual(content.__owner__, None)

        ptah.authService.set_userid('user')
        config.notify(ptah.cms.ContentCreatedEvent(content))

        self.assertEqual(content.__owner__, 'user')

    def test_content_info(self):
        import ptah, ptah.cms

        class MyContent(ptah.cms.Content):
            __mapper_args__ = {'polymorphic_identity': 'mycontent'}
            __uri_factory__ = ptah.UriFactory('mycontent')

        content = MyContent()
        config.notify(ptah.cms.ContentCreatedEvent(content))

        info = content.info()
        self.assertIn('__name__', info)
        self.assertIn('__type__', info)

        class MyContent(ptah.cms.Content):
            __type__ = ptah.cms.Type('mycontent', 'MyContent')

        content = MyContent()
        config.notify(ptah.cms.ContentCreatedEvent(content))

        info = content.info()
        self.assertIn('title', info)
        self.assertIn('description', info)

    def test_content_update(self):
        import ptah, ptah.cms

        class MyContent(ptah.cms.Content):
            __type__ = ptah.cms.Type('mycontent', 'MyContent')

        content = MyContent()
        config.notify(ptah.cms.ContentCreatedEvent(content))

        modified = content.modified
        time.sleep(0.1)

        content.update(title='Test title')
        info = content.info()

        self.assertEqual(info['title'], 'Test title')
        self.assertEqual(content.title, 'Test title')
        self.assertTrue(content.modified > modified)

    def test_content_delete(self):
        import ptah, ptah.cms

        class MyContent(ptah.cms.Content):
            __type__ = ptah.cms.Type('mycontent', 'MyContent')

        class MyContainer(ptah.cms.Container):
            __type__ = ptah.cms.Type('container', 'Container')

        content = MyContent()

        self.assertRaises(ptah.cms.Error, content.delete)

        container = MyContainer()
        container['content'] = content

        content.delete()
        self.assertEqual(container.keys(), [])



""" static resource management api """
import os, re, colander
from urlparse import urlparse

from pyramid.static import _FileResponse
from pyramid.httpexceptions import HTTPNotFound

from ptah import config
from ptah.view import tmpl
from ptah.view.customize import LayerWrapper


STATIC = config.register_settings(
    'static',

    config.SchemaNode(
        colander.Str(),
        name = 'url',
        default = 'static'),

    config.SchemaNode(
        colander.Int(),
        name = 'cache_max_age',
        default = 0),

    title = 'Static resources management',
)

STATIC_ID = 'ptah.view:static'


def static(name, path, layer=''):
    """ Register new static assets directory

    :param name: Resource name
    :param path: Filesystem path
    """
    abspath, pkg = tmpl.path(path)

    if not abspath:
        raise ValueError("Can't find path to static resource")

    if not os.path.isdir(abspath):
        raise ValueError("path is not directory")

    discriminator = (STATIC_ID, name, layer)

    info = config.DirectiveInfo()
    info.attach(
        config.Action(
            LayerWrapper(lambda cfg, a1,a2,a3: \
                         cfg.get_cfg_storage(STATIC_ID)\
                             .update({a1: (a2, a3)}), discriminator),
            (name, abspath, pkg),
            discriminator = discriminator))


def static_url(request, name, path='', **kw):
    url = STATIC.url

    if STATIC.get('isurl', False):
        return '%s/%s/%s'%(url, name, path)
    else:
        url = request.route_url('%s-%s'%(url, name), subpath = path)
        if not path:
            return url[:-1]
        return url


def buildTree(path, not_allowed=re.compile('^[.~$#]')):
    data = {}

    for item in os.listdir(path):
        if not_allowed.match(item):
            continue

        _bp = os.path.join(path, item)
        if os.path.isdir(_bp):
            d = buildTree(os.path.join(path, item), not_allowed)
            for sitem, _t in d.items():
                data[os.path.join(item, sitem)] = 1
        else:
            data[item] = 1

    return data


@config.subscriber(config.AppStarting)
def initialize(ev):
    url = STATIC.url
    if not urlparse(url)[0]:
        registry = config.get_cfg_storage(STATIC_ID)
        for name, (abspath, pkg) in registry.items():
            prefix = '%s/%s'%(url, name)
            rname = '%s-%s'%(url, name)
            pattern = '%s/*subpath'%prefix

            ev.config.add_route(rname, pattern)
            ev.config.add_view(
                route_name =rname,
                view = StaticView(abspath, prefix))
    else:
        STATIC['isurl'] = True


class StaticView(object):

    def __init__(self, path, prefix):
        self.path = path
        self.prefix = '/%s'%prefix
        self.lprefix = len(self.prefix)+1

    def __call__(self, context, request):
        path_info = request.environ.get('PATH_INFO')
        if not path_info:
            return HTTPNotFound(request.url)

        # windows only
        path = os.path.join(
            self.path, os.path.join(*(path_info[self.lprefix:].split('/'))))

        if os.path.isfile(path):
            return _FileResponse(path, STATIC.cache_max_age)

        return HTTPNotFound(request.url)



import ptah
from ptah import form, config
from ptah.form.field import PREVIEW_ID
from ptah.manage import fieldpreviews
from pyramid.testing import DummyRequest

from base import Base


class TestFieldPreviews(Base):

    def test_multiChoicePreview(self):
        previews = config.get_cfg_storage(PREVIEW_ID)
        self.assertIs(previews[form.MultiChoiceField],
                      fieldpreviews.multiChoicePreview)

        request = DummyRequest()
        html = fieldpreviews.multiChoicePreview(request)
        self.assertIn('Multi choice field', html)

    def test_choicePreview(self):
        previews = config.get_cfg_storage(PREVIEW_ID)
        self.assertIs(previews[form.ChoiceField],
                      fieldpreviews.choicePreview)

        request = DummyRequest()
        html = fieldpreviews.choicePreview(request)
        self.assertIn('Choice field', html)

    def test_boolPreview(self):
        previews = config.get_cfg_storage(PREVIEW_ID)
        self.assertIs(previews[form.BoolField],
                      fieldpreviews.boolPreview)

        request = DummyRequest()
        html = fieldpreviews.boolPreview(request)
        self.assertIn('Boolean field', html)

    def test_radioPreview(self):
        previews = config.get_cfg_storage(PREVIEW_ID)
        self.assertIs(previews[form.RadioField],
                      fieldpreviews.radioPreview)

        request = DummyRequest()
        html = fieldpreviews.radioPreview(request)
        self.assertIn('Radio field', html)

    def test_textareaPreview(self):
        previews = config.get_cfg_storage(PREVIEW_ID)
        self.assertIs(previews[form.TextAreaField],
                      fieldpreviews.textareaPreview)

        request = DummyRequest()
        html = fieldpreviews.textareaPreview(request)
        self.assertIn('TextArea field', html)

    def test_linesPreview(self):
        previews = config.get_cfg_storage(PREVIEW_ID)
        self.assertIs(previews[form.LinesField],
                      fieldpreviews.linesPreview)

        request = DummyRequest()
        html = fieldpreviews.linesPreview(request)
        self.assertIn('Lines field', html)

    def test_textPreview(self):
        previews = config.get_cfg_storage(PREVIEW_ID)
        self.assertIs(previews[form.TextField],
                      fieldpreviews.textPreview)

        request = DummyRequest()
        html = fieldpreviews.textPreview(request)
        self.assertIn('Text field', html)

    def test_intPreview(self):
        previews = config.get_cfg_storage(PREVIEW_ID)
        self.assertIs(previews[form.IntegerField],
                      fieldpreviews.intPreview)

        request = DummyRequest()
        html = fieldpreviews.intPreview(request)
        self.assertIn('Integer field', html)

    def test_floatPreview(self):
        previews = config.get_cfg_storage(PREVIEW_ID)
        self.assertIs(previews[form.FloatField],
                      fieldpreviews.floatPreview)

        request = DummyRequest()
        html = fieldpreviews.floatPreview(request)
        self.assertIn('Float field', html)

    def test_decimalPreview(self):
        previews = config.get_cfg_storage(PREVIEW_ID)
        self.assertIs(previews[form.DecimalField],
                      fieldpreviews.decimalPreview)

        request = DummyRequest()
        html = fieldpreviews.decimalPreview(request)
        self.assertIn('Decimal field', html)

    def test_passwordPreview(self):
        previews = config.get_cfg_storage(PREVIEW_ID)
        self.assertIs(previews[form.PasswordField],
                      fieldpreviews.passwordPreview)

        request = DummyRequest()
        html = fieldpreviews.passwordPreview(request)
        self.assertIn('Password field', html)


class TestFieldsModule(Base):

    def test_fields_module(self):
        from ptah.manage.manage import CONFIG, PtahManageRoute
        from ptah.manage.fields import FieldsModule

        request = DummyRequest()

        ptah.authService.set_userid('test')
        CONFIG['managers'] = ('*',)
        mr = PtahManageRoute(request)
        mod = mr['fields']

        self.assertIsInstance(mod, FieldsModule)

    def test_fields_view(self):
        from ptah.manage.fields import FieldsModule, FieldsView

        request = DummyRequest()

        mod = FieldsModule(None, request)

        res = FieldsView.__renderer__(mod, request)
        self.assertEqual(res.status, '200 OK')

        from ptah.form.field import FIELD_ID

        fields = config.get_cfg_storage(FIELD_ID)

        view = FieldsView(None, request)
        view.update()

        self.assertEqual(len(view.fields), len(fields))
        self.assertIn('name', view.fields[0])
        self.assertIn('preview', view.fields[0])



from paste.httpserver import serve
import sqlalchemy as sqla
from pyramid.request import Request

import ptah
from ptah import cms, config, view


FOO = cms.ApplicationFactory(
    '/foo', 'foo', 'Foo Application')

view.register_route(
    'foo', '/foo/*traverse', factory = FOO, use_global_views = True)

if __name__ == '__main__':
    app = ptah.make_wsgi_app({'settings':r'./ptah.ini'})
    request = Request.blank('/') # all pyramid requests have a root
    foo = FOO()
    request.root = foo
    request.registry = app.registry

    from ptah.cmsapp.content import Page

    # Only using the API all events will notify
    if 'page1' not in foo:
        page1 = foo.create(Page.__type__.__uri__, 'page1',
                           text='<p> some html</p>')
        print'page1', foo['page1'], request.resource_url(page1)

    # We can use SQLAlchemy and Container API, which will notify
    if 'page2' not in foo:
        page2 = Page(text='</p>page 2 html</p>')
        cms.Session.add(page2)
        foo['page2'] = page2
        print 'page2', foo['page2']

    # We can just use SQLAlchemy and manually notify application
    #if not cms.Session.query(Page).filter_by(__name__='page3').all():
    if 'page3' not in foo:
        page3 = Page(text='<p>page 3 html</p>',
                     __name__ = 'page3',
                     __parent__ = foo)
        cms.Session.add(page3)
        config.notify(cms.events.ContentCreatedEvent(page3))
        import transaction; transaction.commit()

    print 'See foo Application, http://localhost:8080/foo/'
    serve(app, '0.0.0.0')



""" introspect module """
import inspect, sys
from pyramid.interfaces import IRoutesMapper, IRouteRequest
from pyramid.interfaces import IViewClassifier, IExceptionViewClassifier

from zope import interface
from zope.interface.interface import InterfaceClass

import ptah
from ptah import config, view, manage, form
from ptah.config import directives
from ptah.config.api import exclude, list_packages

from manage import INTROSPECT_ID


class IntrospectModule(manage.PtahModule):
    __doc__ = u'Insight into all configuration and registrations.'

    title = 'Introspect'
    manage.module('introspect')

    packages = None

    def list_packages(self):
        if self.packages is None:
            self.packages = list_packages()
            self.packages.sort()
        return self.packages

    def __getitem__(self, key):
        packages = self.list_packages()
        if key in packages:
            return Package(key, self, self.request)

        raise KeyError(key)


class Package(object):

    def __init__(self, pkg, mod, request):
        self.__name__ = pkg
        self.__parent__ = mod

        self.pkg = pkg
        self.request = request

    def actions(self):
        actions = directives.scan(self.pkg, set(), exclude)

        info = {}

        for action in actions:
            d = action.discriminator[0]
            d_data = info.setdefault(d, {})
            mod_data = d_data.setdefault(action.info.module.__name__, [])
            mod_data.append(action)

        return info


class MainView(view.View):
    view.pview(
        context = IntrospectModule,
        template = view.template('ptah.manage:templates/introspect.pt'))

    __doc__ = 'Introspection module view.'
    __intr_path__ = '/ptah-manage/introspect/index.html'

    def update(self):
        self.packages = self.context.list_packages()


view.register_snippet(
    'ptah-module-actions', IntrospectModule,
    template = view.template('ptah.manage:templates/introspect-actions.pt'))


class PackageView(view.View):
    view.pview(
        context = Package,
        template = view.template('ptah.manage:templates/introspect-pkg.pt'))

    __doc__ = 'Package introspection page.'
    __intr_path__ = '/ptah-manage/introspect/${pkg}/index.html'

    def update(self):
        self.data = self.context.actions()

        self.ndata = ndata = {}
        for tp, d in self.data.items():
            actions = []
            for k, ac in d.items():
                actions.extend(ac)

            ndata[tp] = actions

        itypes = []
        intros = config.get_cfg_storage(INTROSPECT_ID)
        for key, cls in intros.items():
            if key in self.data:
                itypes.append((cls.title, cls(self.request)))

        itypes.sort()
        self.itypes = [it for _t, it in itypes]


def lineno(ob):
    if ob is not None:
        return inspect.getsourcelines(ob)[-1]


class EventsView(view.View):
    view.pview(
        'events.html', IntrospectModule,
        template = view.template('ptah.manage:templates/introspect-events.pt'))

    __doc__ = 'Events introspection page.'
    __intr_path__ = '/ptah-manage/introspect/events.html'

    events = None
    actions = None
    lineno = staticmethod(lineno)

    def update(self):
        ev = self.request.params.get('ev')

        all_events = config.get_cfg_storage(directives.EVENT_ID)
        self.event = event = all_events.get(ev)

        if event is None:
            events = []
            for n, ev in all_events.items():
                if isinstance(n, basestring):
                    events.append((ev.title, ev))

            events.sort()
            self.events = [ev for _t, ev in events]
        else:
            pkgs = list_packages()
            evinst = event.instance

            seen = set()
            actions = []
            for pkg in pkgs:
                for action in directives.scan(pkg, seen, exclude):
                    if action.discriminator[0] == 'ptah.config:subscriber':
                        required = action.args[2]
                        if len(required) == 2 and required[1] == evinst:
                            actions.append(action)
                        elif required[0] == evinst:
                            actions.append(action)

            self.actions = actions


class RoutesView(view.View):
    view.pview(
        'routes.html', IntrospectModule,
        template = view.template('ptah.manage:templates/introspect-routes.pt'))

    __doc__ = 'Routes introspection page.'
    __intr_path__ = '/ptah-manage/introspect/routes.html'

    def update(self):
        self.route = route = None

        if route is None:
            packages = list_packages()

            viewactions = []

            seen = set()
            routes = {}
            for pkg in packages:
                actions = directives.scan(pkg, seen, exclude)

                for action in actions:
                    d = action.discriminator[0]
                    if d == 'ptah.view:route':
                        name, pattern, factory = action.args[:3]
                        routes[name] = (pattern, name, factory, [])
                    elif d == 'ptah.view:view':
                        factory = action.info.context
                        if inspect.isclass(factory):
                            isclass = True
                            name = action.args[0]
                            context = action.args[1]
                            route = action.args[3]
                        else:
                            factory = action.args[0]
                            name = action.args[1]
                            context = action.args[2]
                            route = action.args[4]
                        if route:
                            viewactions.append(
                                (route, name, context, factory, action))

            sm = self.request.registry

            # add pyramid routes
            for route in sm.getUtility(IRoutesMapper).get_routes():
                if route.name not in routes:
                    routes[route.name] = (
                        route.pattern, route.name, route.factory, [])

            # attach views to routes
            for route, name, context, factory, action in viewactions:
                try:
                    rdata = routes[route][3]
                except: # pragma: no cover
                    continue
                rdata.append([getattr(factory, '__intr_path__', name),
                              action.info.module.__name__, lineno(factory),
                              factory, action.discriminator[-1]])
                rdata.sort()

            routes = routes.values()
            routes.sort()
            self.routes = routes

            # views
            route_requests = [i for n, i in sm.getUtilitiesFor(IRouteRequest)]

            views = []
            data = sm.adapters._adapters[3]
            for classifier, data in data.items():
                if classifier in (IViewClassifier, IExceptionViewClassifier):
                    for req, data2 in data.items():
                        if req in route_requests:
                            continue

                        for context, data3 in data2.items():
                            if isinstance(context, InterfaceClass):
                                context = '%s.%s'%(
                                    context.__module__, context.__name__)
                            else:
                                context = context.__name__

                            for provides, adapters in data3.items():
                                for name, factory in adapters.items():
                                    views.append(
                                        (context,name,classifier,req,factory))

            views.sort()
            self.views = views


class EventDirective(object):
    """ zca event declarations """

    title = 'Events'
    manage.introspection('ptah.config:event')

    actions = view.template('ptah.manage:templates/directive-event.pt')

    def __init__(self, request):
        self.request = request

    def renderActions(self, *actions):
        return self.actions(
            actions = actions,
            events = config.get_cfg_storage(directives.EVENT_ID),
            manage_url = manage.CONFIG.manage_url,
            request = self.request)


class AdapterDirective(object):
    """ zc adapter registrations """

    title = 'zc adapters'
    manage.introspection('ptah.config:adapter')

    actions = view.template('ptah.manage:templates/directive-adapter.pt')

    def __init__(self, request):
        self.request = request

    def getInfo(self, action):
        context = action.info.context

        if inspect.isclass(context):
            isclass = True
            requires, name = action.args[:2]
        else:
            context = action.args[1]
            requires = action.args[2]
            name = action.kw['name']

        provided = list(interface.implementedBy(context))
        if len(provided):
            iface = provided[0]
        else: # pragma: no cover
            iface = 'unknown'

        return {'iface': iface,
                'context': context,
                'provided': provided,
                'name': name,
                'requires': requires}

    def renderActions(self, *actions):
        return self.actions(
            actions = actions,
            getInfo = self.getInfo,
            manage_url = manage.CONFIG.manage_url,
            request = self.request)


class SnippetTypeDirective(object):
    """ ptah snippet types """

    title = 'Snippet Types'
    manage.introspection('ptah.view:snippettype')

    actions = view.template('ptah.manage:templates/directive-stype.pt')

    def __init__(self, request):
        self.request = request

    def renderActions(self, *actions):
        STYPE_ID = sys.modules['ptah.view.snippet'].STYPE_ID
        stypes = config.get_cfg_storage(STYPE_ID)
        return self.actions(
            actions = actions,
            stypes = stypes,
            manage_url = manage.CONFIG.manage_url,
            request = self.request)


class RouteDirective(object):
    """ pyramid routes """

    title = 'Routes'
    manage.introspection('ptah.view:route')

    actions = view.template('ptah.manage:templates/directive-route.pt')

    def __init__(self, request):
        self.request = request

    def renderActions(self, *actions):
        return self.actions(
            actions = actions,
            manage_url = manage.CONFIG.manage_url,
            request = self.request)


class SubscriberDirective(object):
    """ zca event subscribers """

    title = 'Event subscribers'
    manage.introspection('ptah.config:subscriber')

    actions = view.template('ptah.manage:templates/directive-subscriber.pt')

    def __init__(self, request):
        self.request = request

    def getInfo(self, action):
        factory, ifaces = action.args[1:]
        factoryInfo = '%s.%s'%(action.info.module.__name__, factory.__name__)

        if len(action.args[2]) > 1:
            obj = action.args[2][0]
            klass = action.args[2][-1]
            event = config.get_cfg_storage(directives.EVENT_ID).get(
                action.args[2][-1], None)
        else:
            obj = None
            klass = action.args[2][0]
            event = config.get_cfg_storage(directives.EVENT_ID).get(
                action.args[2][0], None)

        return locals()

    def renderActions(self, *actions):
        return self.actions(
            getInfo = self.getInfo,
            actions = actions,
            manage_url = manage.CONFIG.manage_url,
            request = self.request)


class ViewDirective(object):
    """ pyramid views """

    title = 'Views'
    manage.introspection('ptah.view:view')

    actions = view.template('ptah.manage:templates/directive-view.pt')

    def __init__(self, request):
        self.request = request

    def getInfo(self, action):
        info = action.info
        factory = action.info.context

        if inspect.isclass(factory):
            isclass = True
            name,context,template,route,layout,permission = action.args
        else:
            isclass = False
            factory,name,context,template,route,layout,permission = action.args

        if route:
            if name:
                view = 'view: "%s" route: "%s"'%((name or '<default>'), route)
            else:
                view = 'route: "%s"'%route
        else:
            view = 'view: %s'%(name or '<default>')

        if isclass:
            factoryInfo = '%s.%s'%(factory.__module__, factory.__name__)
        else:
            factoryInfo = '%s.%s'%(info.module.__name__, factory.__name__)

        if template:
            template = template.spec
        else:
            template = ''

        return locals()

    def renderActions(self, *actions):
        return self.actions(
            getInfo = self.getInfo,
            actions = actions,
            manage_url = manage.CONFIG.manage_url,
            request = self.request)



from ptah import config, view
from collections import OrderedDict
from pyramid.location import lineage
from pyramid.security import ACLDenied, Allow, Deny
from pyramid.security import ALL_PERMISSIONS, NO_PERMISSION_REQUIRED
from pyramid.interfaces import IAuthorizationPolicy
from pyramid.threadlocal import get_current_registry
from pyramid.httpexceptions import HTTPForbidden

from ptah import authService
from ptah import SUPERUSER_URI
from ptah.interfaces import IOwnersAware
from ptah.interfaces import ILocalRolesAware


ACL_ID = 'ptah:acl-map'
ROLE_ID = 'ptah:role'
PERMISSION_ID = 'ptah:permission'


def get_acls():
    return config.get_cfg_storage(ACL_ID)


def get_roles():
    return config.get_cfg_storage(ROLE_ID)


def get_permissions():
    return config.get_cfg_storage(PERMISSION_ID)


class PermissionInfo(str):

    title = u''
    description = u''


def Permission(name, title, description=u''):
    """ Register new permission. """
    info = config.DirectiveInfo()

    permission = PermissionInfo(name)
    permission.title = title
    permission.description = description

    info.attach(
        config.Action(
            lambda config, p: \
                config.get_cfg_storage(PERMISSION_ID).update({str(p): p}),
            (permission,),
            discriminator=(PERMISSION_ID, name))
        )

    return permission


class ACL(list):
    """ Named ACL map

    ACL contains list of permit rules, for example::

      >> acl = ACL('test', 'Test ACL')
      >> acl.allow('system.Everyone', 'View')
      >> acl.deny('system.Everyone', 'Edit')

      >> list(acl)
      [(Allow, 'system.Everyone', ('View',)),
       (Deny, 'system.Everyone', ('Edit',))]

    """

    # do we need somthing like Unset, to unset permission from parent

    def __init__(self, name, title, description=u''):
        self.name = name
        self.title = title
        self.description = description

        info = config.DirectiveInfo()
        info.attach(
            config.Action(
                lambda config, p: \
                    config.get_cfg_storage(ACL_ID).update({name: p}),
                (self,),
                discriminator=('ptah:acl-map', name))
            )
        self.directiveInfo = info

    def get(self, typ, role):
        for r in self:
            if r[0] == typ and r[1] == role:
                return r

        return None

    def allow(self, role, *permissions):
        """ Give permissions to role """

        if not isinstance(role, basestring):
            role = role.id

        rec = self.get(Allow, role)
        if rec is None:
            rec = [Allow, role, set()]
            self.append(rec)

        if rec[2] is ALL_PERMISSIONS:
            return

        if ALL_PERMISSIONS in permissions:
            rec[2] = ALL_PERMISSIONS
        else:
            rec[2].update(permissions)

    def deny(self, role, *permissions):
        """ Deny permissions for role """

        if not isinstance(role, basestring):
            role = role.id

        rec = self.get(Deny, role)
        if rec is None:
            rec = [Deny, role, set()]
            self.append(rec)

        if rec[2] is ALL_PERMISSIONS:
            return

        if ALL_PERMISSIONS in permissions:
            rec[2] = ALL_PERMISSIONS
        else:
            rec[2].update(permissions)

    def unset(self, role, *permissions):
        """ Unset any previously defined permissions """
        for perm in permissions:
            for rec in self:
                if role is not None and rec[1] != role:
                    continue

                if rec[2] is ALL_PERMISSIONS or perm is ALL_PERMISSIONS:
                    rec[2] = set()
                else:
                    if perm in rec[2]:
                        rec[2].remove(perm)

        records = []
        for rec in self:
            if rec[2]:
                records.append(rec)
        self[:] = records


class ACLsMerge(object):

    def __init__(self, acls):
        self.acls = acls

    def __iter__(self):
        acls = config.get_cfg_storage(ACL_ID)
        for aname in self.acls:
            acl = acls.get(aname)
            if acl is not None:
                for rec in acl:
                    yield rec


class ACLsProperty(object):
    """ This property merges `__acls__` list of ACLs and
    generate one `__acl__`

    For example::

      >> class Content(object):
      ...
      ...   __acls__ = ['map1', 'map2']
      ...
      ...   __acl__ = ACLsProperty()

    In this case it is possible to manipulate permissions
    by just changing `__acls__` list.

    """

    def __get__(self, inst, klass):
        acls = getattr(inst, '__acls__', ())
        if acls:
            return ACLsMerge(acls)
        else:
            return ()


class Role(object):
    """ Register new security role in the system """

    def __init__(self, name, title, description='',
                 prefix='role:', system=False):
        id = '%s%s' % (prefix, name)

        self.id = id
        self.name = name
        self.title = title
        self.description = description
        self.system = system

        self.allowed = set()
        self.denied = set()

        # conflict detection and introspection
        info = config.DirectiveInfo()
        info.attach(
            config.Action(
                lambda config, r: \
                    config.get_cfg_storage(ROLE_ID).update({r.name: r}),
                (self, ),
                discriminator=(ROLE_ID, name))
            )

    def __str__(self):
        return 'Role<%s>' % self.title

    def __repr__(self):
        return self.id

    def allow(self, *permissions):
        DEFAULT_ACL.allow(self.id, *permissions)

    def deny(self, *permissions):
        DEFAULT_ACL.deny(self.id, *permissions)

    def unset(self, *permissions):
        DEFAULT_ACL.unset(self.id, *permissions)


def get_local_roles(userid, request=None, context=None):
    """ calculates local roles for userid """
    if context is None:
        context = getattr(request, 'context', None)
        if context is None:
            context = getattr(request, 'root', None)

    roles = OrderedDict()

    if IOwnersAware.providedBy(context):
        if userid == context.__owner__:
            roles[Owner.id] = Allow

    for location in lineage(context):
        if ILocalRolesAware.providedBy(location):
            local_roles = location.__local_roles__
            if local_roles:
                for r in local_roles.get(userid, ()):
                    if r not in roles:
                        roles[r] = Allow

    data = []
    for r, val in roles.items():
        if val is Allow:
            data.append(r)

    return data


DEFAULT_ACL = ACL('', 'Default ACL map')

Everyone = Role(
    'Everyone', 'Everyone', '', 'system.', True)

Authenticated = Role(
    'Authenticated', 'Authenticated', '', 'system.', True)

Owner = Role(
    'Owner', 'Owner', '', 'system.', True)

NOT_ALLOWED = Permission('__not_allowed__', 'Special permission')


def check_permission(permission, context, request=None, throw=False):
    """ Check `permission` withing `context`.

    :param permission: Permission
    :type permission: (Permission or sting)
    :param context: Context object
    :param throw: Throw HTTPForbidden exception.
    """

    if not permission or permission == NO_PERMISSION_REQUIRED:
        return True
    if permission == NOT_ALLOWED:
        if throw:
            raise HTTPForbidden()
        return False

    userid = authService.get_userid()
    if userid == SUPERUSER_URI:
        return True

    AUTHZ = get_current_registry().getUtility(IAuthorizationPolicy)

    principals = [Everyone.id]

    if userid is not None:
        principals.extend((Authenticated.id, userid))

        roles = get_local_roles(userid, context=context)
        if roles:
            principals.extend(roles)

    res = AUTHZ.permits(context, principals, permission)

    if isinstance(res, ACLDenied):
        if throw:
            raise HTTPForbidden(res)

        return False
    return True

view.set_checkpermission(check_permission)


@config.cleanup
def cleanup():
    DEFAULT_ACL[:] = []



""" custom SchemaNode, Mapping, Sequence class """
import colander


class Required(colander.Invalid):

    def __init__(self, node, msg='Required', value=None):
        super(Required, self).__init__(node, msg, value)


class RequiredWithDependency(object):
    """ Validator which fails only if other key is set. """

    def __init__(self, key, depkey,
                 depval=colander.required, default=colander.required):
        self.key = key
        self.depkey = depkey
        self.depval = depval
        self.default = default

    def __call__(self, node, appstruct):
        if appstruct:
            val = appstruct.get(self.depkey)

            hasVal = False
            if self.depval is not colander.required:
                hasVal = val == self.depval
            else:
                if val:
                    hasVal = True

            if hasVal:
                node = node[self.key]
                default = self.default
                if default is colander.required:
                    default = node.default

                val = appstruct.get(self.key, default)
                if val == default:
                    raise Required(node)


class Mapping(colander.Mapping):

    def flatten(self, node, appstruct, prefix='', listitem=False):
        result = {}
        selfprefix = ''
        if listitem:
            selfprefix = prefix
        elif node.name:
            selfprefix = '%s%s.' % (prefix, node.name)

        for subnode in node.children:
            name = subnode.name
            if name in appstruct:
                substruct = appstruct[name]
                result.update(subnode.typ.flatten(
                        subnode, substruct, prefix=selfprefix))
        return result

    def unflatten(self, node, paths, fstruct):
        return _unflatten_mapping(node, paths, fstruct)


class Sequence(colander.Sequence):

    def flatten(self, node, appstruct, prefix='', listitem=False):
        result = {}
        selfprefix = ''
        if listitem:
            selfprefix = prefix
        elif node.name:
            selfprefix = '%s%s.' % (prefix, node.name)

        childnode = node.children[0]

        for num, subval in enumerate(appstruct):
            subname = '%s%s' % (selfprefix, num)
            subprefix = subname
            result.update(childnode.typ.flatten(
                    childnode, subval, prefix=subprefix, listitem=True))

        return result


Seq = Sequence


class String(colander.Str):

    def __init__(self, encoding='utf-8'):
        self.encoding = encoding


Str = String
colander.Str = String
colander.String = String


class SchemaNode(colander.SchemaNode):

    def __init__(self, typ, *children, **kw):
        super(SchemaNode, self).__init__(typ, *children, **kw)

        self.required = kw.pop('required', self.default is colander.null)
        self._origin_default = self.default

    def __get_required(self):
        return self.__dict__.get('required')

    def __set_required(self, val):
        self.__dict__['required'] = val

    required = property(__get_required, __set_required)

    def deserialize(self, cstruct=colander.null):
        appstruct = self.typ.deserialize(self, cstruct)
        if appstruct is colander.null:
            appstruct = self.default

        if self.preparer is not None:
            appstruct = self.preparer(appstruct)

        if self.required and appstruct == self.default:
            raise Required(self)

        if self.validator is not None:
            self.validator(self, appstruct)

        return appstruct


def _unflatten_mapping(node, paths, fstruct,
                       get_child=None, rewrite_subpath=None):
    if get_child is None:
        get_child = node.__getitem__
    if rewrite_subpath is None:
        def rewrite_subpath(subpath):
            return subpath
    node_name = node.name
    if node_name:
        prefix = node_name + '.'
    else:
        prefix = ''
    prefix_len = len(prefix)
    appstruct = {}
    subfstruct = {}
    subpaths = []
    curname = None
    for path in paths:
        if path == node_name or (prefix and not path.startswith(prefix)):
            # flattened structs contain non-leaf nodes which are ignored
            # during unflattening.
            continue  # pragma: no cover
        assert path.startswith(prefix), "Bad node: %s" % path
        subpath = path[prefix_len:]
        if '.' in subpath:
            name = subpath[:subpath.index('.')]
        else:
            name = subpath
        if curname is None:
            curname = name
        elif name != curname:
            try:
                subnode = get_child(curname)
            except:
                subfstruct = {}
                subpaths = []
                curname = name
            else:
                appstruct[curname] = subnode.typ.unflatten(
                    subnode, subpaths, subfstruct)
                subfstruct = {}
                subpaths = []
                curname = name

        subpath = rewrite_subpath(subpath)
        subfstruct[subpath] = fstruct[path]
        subpaths.append(subpath)

    if curname is not None:
        try:
            subnode = get_child(curname)
        except KeyError:
            pass
        else:
            appstruct[curname] = subnode.typ.unflatten(
                subnode, subpaths, subfstruct)

    return appstruct



from ptah import config, view

from base import Base


class TestLibraryManagement(Base):

    def _setup_ptah(self):
        pass

    def tearDown(self):
        config.cleanup_system(self.__class__.__module__)
        super(TestLibraryManagement, self).tearDown()

    def test_library_register_fail(self):
        # path resuired
        self.assertRaises(ValueError, view.library, 'test')

        # resource should be registered
        self.assertRaises(ValueError, view.library,
                          'test', path='/test.js', resource='unknown')

        # type should be 'js' or 'css'
        self.assertRaises(ValueError, view.library,
                          'test', path='/test.js', type='unknown')

        # if not resource is provided, path has to be absolute url
        self.assertRaises(ValueError, view.library,
                          'test', path='/test.js', type='js')

    def test_library_simple_css(self):
        view.static('tests', 'ptah.view.tests:static/dir1')

        view.library(
            'test-lib', path='style.css', resource='tests', type='css')
        self._init_ptah()

        from ptah.view.library import LIBRARY_ID
        lib = config.get_cfg_storage(LIBRARY_ID)['test-lib']

        self.assertEqual(lib.name, 'test-lib')
        self.assertEqual(len(lib.entries), 1)
        self.assertEqual(len(lib.entries), 1)
        self.assertTrue('style.css' in lib.entries[0].paths)

        self.assertEqual(
            repr(lib), '<ptah.view.library.Library "test-lib">')

    def test_library_simple_js(self):
        view.library(
            'test-lib', path='http://ptah.org/test.js', type='js')
        self._init_ptah()
        from ptah.view.library import LIBRARY_ID
        lib = config.get_cfg_storage(LIBRARY_ID)['test-lib']

        self.assertEqual(
            lib.render(self.request),
            '<script src="http://ptah.org/test.js"> </script>')

    def test_library_render_absurls(self):
        view.library(
            'test-lib', path='http://ptah.org/style.css', type='css')
        self._init_ptah()
        from ptah.view.library import LIBRARY_ID
        lib = config.get_cfg_storage(LIBRARY_ID)['test-lib']

        self.assertEqual(
            lib.render(self.request),
            '<link type="text/css" rel="stylesheet" href="http://ptah.org/style.css" />')

    def test_library_render_with_prefix_postfix(self):
        view.library(
            'test-lib', path='http://ptah.org/style.css', type='css',
            prefix='<!--[if lt IE 7 ]>', postfix='<![endif]-->')
        self._init_ptah()
        from ptah.view.library import LIBRARY_ID
        lib = config.get_cfg_storage(LIBRARY_ID)['test-lib']

        self.assertEqual(
            lib.render(self.request),
            '<!--[if lt IE 7 ]><link type="text/css" rel="stylesheet" href="http://ptah.org/style.css" /><![endif]-->')

    def test_library_render_with_extra(self):
        view.library(
            'test-lib', path='http://ptah.org/test.js', type='js',
            extra={'test': "extra"})
        self._init_ptah()
        from ptah.view.library import LIBRARY_ID
        lib = config.get_cfg_storage(LIBRARY_ID)['test-lib']

        self.assertEqual(
            lib.render(self.request),
            '<script test="extra" src="http://ptah.org/test.js"> </script>')

    def test_library_include(self):
        view.library(
            'test-lib', path='http://ptah.org/style.css', type='css')
        self._init_ptah()

        view.include(self.request, 'test-lib')
        self.assertEqual(
            view.render_includes(self.request),
            '<link type="text/css" rel="stylesheet" href="http://ptah.org/style.css" />')

    def test_library_include_errors(self):
        # render not included
        self.assertEqual(view.render_includes(self.request), '')

        # include unknown
        view.include(self.request, 'test-lib-test')
        self.assertEqual(view.render_includes(self.request), '')

    def test_library_include_recursive(self):
        view.library(
            'test-lib1', path='http://ptah.org/style1.css', type='css')

        view.library(
            'test-lib2', path='http://ptah.org/style2.css', type='css',
            require='test-lib1')

        view.library(
            'test-lib3', path='http://ptah.org/style3.css', type='css',
            require=('test-lib1', 'test-lib2'))

        view.library(
            'test-lib4', path='http://ptah.org/style4.css', type='css',
            require=('test-lib1', 'test-lib2'))
        self._init_ptah()

        view.include(self.request, 'test-lib3')
        view.include(self.request, 'test-lib4')

        self.assertEqual(
            view.render_includes(self.request),
"""<link type="text/css" rel="stylesheet" href="http://ptah.org/style1.css" />
<link type="text/css" rel="stylesheet" href="http://ptah.org/style2.css" />
<link type="text/css" rel="stylesheet" href="http://ptah.org/style3.css" />
<link type="text/css" rel="stylesheet" href="http://ptah.org/style4.css" />""")

    def test_library_include_resource(self):
        view.static('tests2', 'ptah.view.tests:static/dir1')
        view.library(
            'test-lib', path='style.css', resource='tests2', type='css')
        self._init_ptah()

        request = self._makeRequest()

        view.include(request, 'test-lib')

        self.assertEqual(
            view.render_includes(request),
            '<link type="text/css" rel="stylesheet" href="http://localhost:8080/static/tests2/style.css" />')

    def test_library_View_include(self):
        view.library('test-lib', path='http://ptah.org/test.js', type='js')
        self._init_ptah()

        base = view.View(None, self.request)
        base.include('test-lib')

        self.assertEqual(
            base.render_includes(),
            '<script src="http://ptah.org/test.js"> </script>')



""" static assets """
from ptah import view

view.static('jquery', 'ptah:static/jquery')
view.static('bootstrap', 'ptah:static/bootstrap')
view.static('tiny_mce', 'ptah:static/tiny_mce')

# jQuery library
view.library(
    'jquery',
    path="jquery-1.6.4.min.js",
    resource="jquery",
    type="js")

view.library(
    'jquery-ui',
    path="jquery-ui-1.8.16.min.js",
    type="js",
    resource="jquery",
    require="jquery")

view.library(
    'jquery-ui',
    path="jquery-ui.css",
    resource="jquery",
    type='css')

# Bootstrap css
view.library(
    'bootstrap',
    path="bootstrap-1.4.0.min.css",
    resource="bootstrap",
    type="css")

view.library(
    'bootstrap-js',
    path=('js/bootstrap-alerts.js',
          'js/bootstrap-buttons.js',
          'js/bootstrap-dropdown.js',
          'js/bootstrap-modal.js',
          'js/bootstrap-popover.js',
          'js/bootstrap-scrollspy.js',
          'js/bootstrap-tabs.js',
          'js/bootstrap-twipsy.js'),
    resource="bootstrap",
    type="js",
    require="jquery")

# TinyMCE
view.library(
    "tiny_mce",
    resource="tiny_mce",
    path=('tiny_mce.js', 'jquery.tinymce.js'),
    type="js",
    require='jquery')



""" introspect module """
import urllib
import sqlahelper as psa
from sqlalchemy.orm.mapper import _mapper_registry
from pyramid.decorator import reify
from pyramid.httpexceptions import HTTPFound

import ptah
from ptah import view, form, manage

Session = psa.get_session()
metadata = psa.get_base().metadata


class SQLAModule(manage.PtahModule):
    __doc__ = u'A listing of all tables with ability to view and edit records'

    title = 'SQLAlchemy'
    manage.module('sqla')

    metadata = {}

    def __getitem__(self, key):
        try:
            id, table = key.split('-', 1)
        except:
            raise KeyError(key)

        md = self.metadata[id][0]
        return Table(md.tables[table], self, self.request)

    @classmethod
    def addMetadata(cls, md, id, title=''):
        cls.metadata[id] = [md, title or id.capitalize()]

addMetadata = SQLAModule.addMetadata
addMetadata(metadata, 'psqla', 'Pyramid sqla')


class Table(object):

    def __init__(self, table, mod, request):
        self.__name__ = table.name
        self.__parent__ = mod

        self.table = table
        self.request = request

    def __getitem__(self, key):
        if key == 'add.html':
            raise KeyError(key)

        try:
            return Record(key, self.table, self, self.request)
        except:
            raise KeyError(key)


class Record(object):

    def __init__(self, pid, table, parent, request):
        self.pid = pid
        self.table = table
        self.request = request

        self.__name__ = str(pid)
        self.__parent__ = parent

        self.pname = None
        self.pcolumn = None
        for cl in table.columns:
            if cl.primary_key:
                self.pname = cl.name
                self.pcolumn = cl

        self.data = Session.query(table).filter(
            self.pcolumn == pid).one()


class MainView(view.View):
    view.pview(
        context = SQLAModule,
        template = view.template('ptah.manage:templates/sqla-index.pt'))

    __doc__ = "sqlalchemy tables listing page."
    __intr_path__ = '/ptah-manage/sqla/index.html'

    def printTable(self, table):
        columns = []
        for cl in table.columns:
            kwarg = []
            if cl.key != cl.name:
                kwarg.append('key') # pragma: no cover
            if cl.primary_key:
                kwarg.append('primary_key')
            if not cl.nullable:
                kwarg.append('nullable')
            if cl.default:
                kwarg.append('default')

            columns.append(
                "Column(%s)" % ', '.join(
                    [repr(cl.name)] + [repr(cl.type)] +
                    [repr(x) for x in cl.foreign_keys if x is not None] +
                    [repr(x) for x in cl.constraints] +
                    ["%s=%s" % (k, repr(getattr(cl, k))) for k in kwarg])
                )

        return ("Table(%s)" % repr(table.name), columns)

    def update(self):
        tables = []

        for id, (md, title) in self.context.metadata.items():
            data = []
            for name, table in md.tables.items():
                data.append((name, self.printTable(table)))

            data.sort()
            tables.append((title, id, data))

        tables.sort()
        self.tables = tables


def get_inheritance(table):
    # inheritance
    inherits = []
    mapper = None
    for m, s in _mapper_registry.items():
        if m.local_table is table:
            mapper = m

    curr_mapper = mapper
    while curr_mapper is not None:
        curr_mapper = curr_mapper.inherits
        if curr_mapper is not None and \
                curr_mapper.local_table.name != table.name:
            inherits.append(curr_mapper.local_table.name)

    inherits.reverse()
    return inherits


class TableView(form.Form):
    view.pview(
        context = Table,
        template = view.template('ptah.manage:templates/sqla-table.pt'))

    __doc__ = "List table records."
    __intr_path__ = '/ptah-manage/sqla/${table}/index.html'

    csrf = True
    page = ptah.Pagination(15)

    def update(self):
        table = self.table = self.context.table
        self.primary = None
        self.pcolumn = None
        self.uris = []

        self.inheritance = get_inheritance(table)

        if table.name == 'ptah_nodes' or self.inheritance:
            self.show_actions = False
        else:
            self.show_actions = True

        names = []
        for cl in table.columns:
            names.append(cl.name)
            if cl.primary_key:
                self.primary = cl.name
                self.pcolumn = cl
            if cl.info.get('uri'):
                self.uris.append(len(names)-1)

        super(TableView, self).update()

        request = self.request
        try:
            current = int(request.params.get('batch', None))
            if not current:
                current = 1

            request.session['table-current-batch'] = current
        except:
            current = request.session.get('table-current-batch')
            if not current:
                current = 1

        self.size = Session.query(table).count()
        self.current = current

        self.pages, self.prev, self.next = self.page(self.size, self.current)

        offset, limit = self.page.offset(current)
        self.data = Session.query(table).offset(offset).limit(limit).all()

    def quote(self, val):
        return urllib.quote_plus(val)

    def val(self, val):
        try:
            if isinstance(val, str): # pragma: no cover
                val = unicode(val, 'utf-8', 'ignore')
            elif not isinstance(val, unicode):
                val = str(val)
        except: # pragma: no cover
            val = u"Can't show"
        return val[:100]

    @form.button('Add', actype=form.AC_PRIMARY)
    def add(self):
        raise HTTPFound(location='add.html')

    @form.button('Remove', actype=form.AC_DANGER)
    def remove(self):
        self.validate_csrf_token()

        ids = []
        for id in self.request.POST.getall('rowid'):
            try:
                ids.append(int(id))
            except:
                pass

        if not ids:
            self.message('Please select records for removing.', 'warning')
            return

        self.table.delete(self.pcolumn.in_(ids)).execute()
        self.message('Select records have been removed.')


class EditRecord(form.Form):
    view.pview(context = Record,
               template = view.template('ptah.manage:templates/sqla-edit.pt'))

    __doc__ = "Edit table record."
    __intr_path__ = '/ptah-manage/sqla/${table}/${record}/index.html'

    csrf = True

    @reify
    def label(self):
        return 'record %s'%self.context.__name__

    @reify
    def fields(self):
        return ptah.build_sqla_fieldset(
            [(cl.name, cl) for cl in self.context.table.columns],
            skipPrimaryKey=True)

    def update(self):
        self.inheritance = get_inheritance(self.context.table)
        if self.context.table.name == 'ptah_nodes' or self.inheritance:
            self.show_remove = False
        else:
            self.show_remove = True

        super(EditRecord, self).update()

    def form_content(self):
        data = {}
        for field in self.fields.fields():
            data[field.name] = getattr(
                self.context.data, field.name, field.default)
        return data

    @form.button('Cancel')
    def cancel_handler(self):
        raise HTTPFound(location='..')

    @form.button('Modify', actype=form.AC_PRIMARY)
    def modify_handler(self):
        data, errors = self.extract()

        if errors:
            self.message(errors, 'form-error')
            return

        self.context.table.update(
            self.context.pcolumn == self.context.__name__, data).execute()
        self.message('Table record has been modified.', 'success')
        raise HTTPFound(location='..')

    @form.button('Remove', actype=form.AC_DANGER,
                 condition=lambda form: form.show_remove)
    def remove_handler(self):
        self.validate_csrf_token()

        self.context.table.delete(
            self.context.pcolumn == self.context.__name__).execute()
        self.message('Table record has been removed.')
        raise HTTPFound(location='..')


class AddRecord(form.Form):
    """ Add new table record. """
    view.pview('add.html', Table)

    __intr_path__ = '/ptah-manage/sqla/${table}/add.html'

    csrf = True

    @reify
    def label(self):
        return '%s: new record'%self.context.table.name

    @reify
    def fields(self):
        return ptah.build_sqla_fieldset(
            [(cl.name, cl) for cl in self.context.table.columns],
            skipPrimaryKey = True)

    @form.button('Create', actype=form.AC_PRIMARY)
    def create(self):
        data, errors = self.extract()

        if errors:
            self.message(errors, 'form-error')
            return

        try:
            self.context.table.insert(values = data).execute()
        except Exception, e: # pragma: no cover
            self.message(e, 'error')
            return

        self.message('Table record has been created.', 'success')
        raise HTTPFound(location='.')

    @form.button('Save & Create new', name='createmulti',actype=form.AC_PRIMARY)
    def create_multiple(self):
        data, errors = self.extract()

        if errors:
            self.message(errors, 'form-error')
            return

        try:
            self.context.table.insert(values = data).execute()
        except Exception, e: # pragma: no cover
            self.message(e, 'error')
            return

        self.message('Table record has been created.', 'success')

    @form.button('Back')
    def cancel(self):
        raise HTTPFound(location='.')



""" base view class with access to various api's """
import logging
from zope import interface
from ptah.formatter import format
from ptah.view.resources import static_url
from ptah.view.message import add_message, render_messages
from ptah.view.library import include, render_includes
from ptah.view.interfaces import ISnippet, IPtahView

log = logging.getLogger('ptah.view')


class View(object):
    interface.implements(IPtahView)

    __name__ = ''
    __parent__ = None

    format = format
    template = None

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.__parent__ = context

    def update(self):
        return {}

    def render(self):
        params = self.update()

        if self.template is not None:
            kwargs = {'view': self,
                      'context': self.context,
                      'request': self.request}
            if type(params) is dict:
                kwargs.update(params)
            return self.template(**kwargs)

        return u''

    def include(self, name):
        include(self.request, name)

    def render_includes(self):
        return render_includes(self.request)

    def message(self, msg, type='info'):
        return add_message(self.request, msg, type)

    def render_messages(self):
        return render_messages(self.request)

    def static_url(self, name, path=''):
        return static_url(self.request, name, path)

    def snippet(self, stype, context=None):
        if context is None:
            context = self.context

        request = self.request

        try:
            snippet = request.registry.queryMultiAdapter(
                (context, request), ISnippet, stype)
            if snippet is not None:
                return snippet()
        except Exception, e:
            log.exception(str(e))

        return u''



import os, pkg_resources
from pyramid.path import caller_package
from chameleon.zpt.template import PageTemplateFile
from chameleon.zpt.template import PageTextTemplateFile
from ptah import config


registry = {}

class _Template(object):

    def __init__(self, default, custom=None, spec=''):
        self.default = default
        self.custom = custom
        self.spec = spec

        self.tmpl = default

    def setCustom(self, custom): # pragma: no cover
        if custom is None:
            self.custom = None
            self.tmpl = self.default
        else:
            self.custom = custom
            self.tmpl = custom

    def __call__(self, *args, **kw):
        return self.tmpl(*args, **kw)

    def __repr__(self):
        return repr(self.tmpl)


def template(spec, layer=None, title=None, description=None, nolayer=False):
    abspath, package_name = path(spec, caller_package(2).__name__)
    if not abspath:
        raise ValueError('Missing template asset: %s' % spec)

    if not nolayer:
        if layer is None:
            layer = package_name

        data = registry.setdefault(layer, {})

        filename = os.path.split(abspath)[1]

        if filename in data:
            tmpl = data[filename][3]
        else:
            tmpl = _Template(getRenderer(abspath), spec=spec)
            data[filename] = [abspath,title,description,tmpl,package_name]
    else:
        tmpl = _Template(getRenderer(abspath), spec=spec)

    return tmpl


def getRenderer(path):
    if path.endswith('.txt'):
        return PageTextTemplateFile(path)

    return PageTemplateFile(path)


def path(spec, package_name=None):
    if os.path.exists(spec):
        return spec, caller_package(3).__name__

    try:
        package_name, filename = spec.split(':', 1)
    except ValueError:
        # somehow we were passed a relative pathname
        if package_name is None:
            package_name = caller_package(2).__name__
        filename = spec

    abspath = pkg_resources.resource_filename(package_name, filename)
    if not pkg_resources.resource_exists(package_name, filename):
        return None, package_name
    return abspath, package_name


@config.cleanup
def cleanup():
    registry.clear()



""" process shutdown """
import logging
import signal
import sys

from signal import SIGINT, SIGTERM

handlers = []


def shutdown_handler(handler):
    handlers.append(handler)
    return handler


_shutdown = False
_handler_int = signal.getsignal(SIGINT)
_handler_term = signal.getsignal(SIGTERM)
log = logging.getLogger('ptah.config')


def shutdown():
    global _shutdown

    if not _shutdown:
        #_shutdown = True

        for handler in handlers:
            try:
                handler()
            except:
                #log.exception("Showndown handler: %s"%handler)
                pass

    #handlers[:] = []


def processShutdown(sig, frame):
    shutdown()

    if sig == SIGINT and callable(_handler_int):
        _handler_int(sig, frame)

    if sig == SIGTERM and callable(_handler_term):  # pragma: no cover
        _handler_term(sig, frame)

    if sig == SIGTERM:
        raise sys.exit()


try:
    import mod_wsgi
except ImportError:
    signal.signal(SIGINT, processShutdown)
    signal.signal(SIGTERM, processShutdown)



from ptah import form, config

from base import Base


class TestFieldset(Base):

    def test_decl(self):
        global MyField

        class MyField(form.Field):

            form.field('my-field')

        self._init_ptah()
        self.assertIs(form.get_field_factory('my-field'), MyField)

    def test_decl_register(self):

        class MyField(form.Field):
            """ """

        form.register_field_factory(MyField, 'my-field')

        self._init_ptah()
        self.assertIs(form.get_field_factory('my-field'), MyField)

    def test_decl_conflict(self):
        global MyField, MyField2

        class MyField(form.Field):

            form.field('my-field')


        class MyField2(form.Field):

            form.field('my-field')

        self.assertRaises(config.ConflictError, self._init_ptah)

    def test_decl_preview(self):
        global MyField

        class MyField(form.Field):
            form.field('my-field')

        @form.fieldpreview(MyField)
        def preview(request):
            """ """

        self._init_ptah()

        from ptah.form.field import PREVIEW_ID
        previews = config.get_cfg_storage(PREVIEW_ID)

        self.assertIn(MyField, previews)
        self.assertIs(previews[MyField], preview)
        self.assertIs(form.get_field_preview(MyField), preview)



import ptah
from ptah import config
from pyramid.testing import DummyRequest
from pyramid.httpexceptions import HTTPForbidden

from base import Base


class TestManageModule(Base):

    def setUp(self):
        self._setup_pyramid()

    def tearDown(self):
        config.cleanup_system(self.__class__.__module__)
        super(TestManageModule, self).tearDown()

    def test_manage_module(self):
        from ptah.manage.manage import \
           module, MANAGE_ID, PtahModule, PtahManageRoute,\
           PtahAccessManager, set_access_manager

        global TestModule
        class TestModule(PtahModule):
            """ module description """

            title = 'Test module'
            module('test-module')

        self._init_ptah()

        set_access_manager(PtahAccessManager)

        MODULES = config.get_cfg_storage(MANAGE_ID)
        self.assertIn('test-module', MODULES)

        request = self._makeRequest()

        self.assertRaises(HTTPForbidden, PtahManageRoute, request)

        ptah.authService.set_userid('test-user')

        self.assertRaises(HTTPForbidden, PtahManageRoute, request)

        def accessManager(id):
            return True

        set_access_manager(accessManager)

        route = PtahManageRoute(request)
        mod = route['test-module']
        self.assertIsInstance(mod, TestModule)
        self.assertTrue(mod.available())
        self.assertEqual(mod.__name__, 'test-module')
        self.assertEqual(mod.url(), '/ptah-manage/test-module')

        self.assertRaises(KeyError, route.__getitem__, 'unknown')

    def test_manage_access_manager(self):
        from ptah.manage.manage import CONFIG
        from ptah.manage.manage import PtahAccessManager

        CONFIG['managers'] = ['*']

        self.assertTrue(PtahAccessManager('test:user'))

        CONFIG['managers'] = ['admin@ptahproject.org']

        self.assertFalse(PtahAccessManager('test:user'))

        class Principal(object):
            id = 'test-user'
            login = 'admin@ptahproject.org'

        principal = Principal()

        @ptah.resolver('test')
        def principalResolver(uri):
            return principal

        self._init_ptah()

        CONFIG['managers'] = ['admin@ptahproject.org']
        self.assertTrue(PtahAccessManager('test:user'))

    def test_manage_view(self):
        from ptah.manage.manage import \
            module, PtahModule, PtahManageRoute, ManageView, \
            set_access_manager

        global TestModule
        class TestModule(PtahModule):
            """ module description """

            title = 'Test module'
            module('test-module')

        def accessManager(id):
            return True

        self._init_ptah()

        set_access_manager(accessManager)
        ptah.authService.set_userid('test-user')

        request = self._makeRequest()
        route = PtahManageRoute(request)

        v = ManageView(route, request)
        v.update()

        self.assertIsInstance(v.modules[-1], TestModule)

    def test_manage_traverse(self):
        from ptah.manage.manage import \
            module, PtahModule, PtahManageRoute, set_access_manager

        global TestModule
        class TestModule(PtahModule):
            """ module description """

            title = 'Test module'
            module('test-module')

        def accessManager(id):
            return True

        self._init_ptah()

        set_access_manager(accessManager)
        ptah.authService.set_userid('test-user')

        request = self._makeRequest()
        route = PtahManageRoute(request)

        mod = route['test-module']
        self.assertIsInstance(mod, TestModule)

    def test_manage_disable_modules(self):
        from ptah.manage.manage import CONFIG, \
            module, PtahModule, PtahManageRoute, ManageView, set_access_manager

        global TestModule
        class TestModule(PtahModule):
            """ module description """

            title = 'Test module'
            module('test-module')

        def accessManager(id):
            return True

        self._init_ptah()

        set_access_manager(accessManager)
        ptah.authService.set_userid('test-user')

        CONFIG['disable_modules'] = ('test-module',)

        request = DummyRequest()
        route = PtahManageRoute(request)

        view = ManageView(route, request)
        view.update()

        for mod in view.modules:
            self.assertFalse(isinstance(mod, TestModule))

    def test_manage_disable_modules_traverse(self):
        from ptah.manage.manage import CONFIG, \
            module, PtahModule, PtahManageRoute, set_access_manager

        global TestModule
        class TestModule(PtahModule):
            """ module description """

            title = 'Test module'
            module('test-module')

        def accessManager(id):
            return True

        self._init_ptah()

        set_access_manager(accessManager)
        ptah.authService.set_userid('test-user')

        CONFIG['disable_modules'] = ('test-module',)

        request = DummyRequest()
        route = PtahManageRoute(request)

        self.assertRaises(KeyError, route.__getitem__, 'test-module')

    def test_manage_layout(self):
        from ptah.manage.manage import \
            module, PtahModule, LayoutManage

        global TestModule
        class TestModule(PtahModule):
            """ module description """

            title = 'Test module'
            module('test-module')

        class Content(object):
            __parent__ = None

        self._init_ptah()

        request = self._makeRequest()

        mod = TestModule(None, request)
        content = Content()
        content.__parent__ = mod
        request.context = content

        layout = LayoutManage(mod, request)
        layout.viewcontext = content
        layout.update()

        self.assertIs(layout.module, mod)


class TestInstrospection(Base):

    def setUp(self):
        self._setup_pyramid()

    def tearDown(self):
        config.cleanup_system(self.__class__.__module__)
        super(TestInstrospection, self).tearDown()

    def test_manage_module(self):
        from ptah.manage.manage import INTROSPECT_ID, introspection

        global TestModule
        class TestModule(object):
            """ module description """

            title = 'Test module'
            introspection('test-module')

        self._init_ptah()

        INTROSPECTIONS = config.get_cfg_storage(INTROSPECT_ID)
        self.assertIn('test-module', INTROSPECTIONS)
        self.assertIs(INTROSPECTIONS['test-module'], TestModule)



""" base class """
import unittest
from ptah import config
from pyramid import testing


class Base(unittest.TestCase):

    def _makeRequest(self, environ=None):
        from pyramid.request import Request
        if environ is None:
            environ = self._makeEnviron()
        return Request(environ)

    def _makeEnviron(self, **extras):
        environ = {
            'wsgi.url_scheme':'http',
            'wsgi.version':(1,0),
            'PATH_INFO': '/',
            'SERVER_NAME':'localhost',
            'SERVER_PORT':'8080',
            'REQUEST_METHOD':'GET',
            }
        environ.update(extras)
        return environ

    def _init_ptah(self, settings={}, handler=None, *args, **kw):
        config.initialize(self.p_config, ('ptah', self.__class__.__module__),
                          initsettings = False)
        config.initialize_settings(settings, self.p_config)
        config.start(self.p_config)

    def _setup_pyramid(self):
        self.request = request = testing.DummyRequest()
        request.params = {}
        self.p_config = testing.setUp(request=request)
        self.registry = self.p_config.registry
        self.request.registry = self.p_config.registry

    def _setup_ptah(self):
        config.initialize(self.p_config, ('ptah', self.__class__.__module__))

    def _setRequest(self, request):
        self.request = request
        self.p_config.end()
        self.p_config.begin(request)

    def setUp(self):
        self._setup_pyramid()
        self._setup_ptah()

    def tearDown(self):
        config.shutdown()
        config.cleanup_system(self.__class__.__module__)
        sm = self.p_config
        sm.__init__('base')
        testing.tearDown()


class Context(object):

    def __init__(self, parent=None, name=''):
        self.__name__ = name
        self.__parent__ = parent



import transaction
from cStringIO import StringIO

from base import Base


class TestBlob(Base):

    def test_blob(self):
        import ptah.cms

        blob = ptah.cms.blobStorage.add(StringIO('blob data'))
        self.assertTrue(ptah.cms.IBlob.providedBy(blob))
        self.assertEqual(blob.read(), 'blob data')
        self.assertTrue(ptah.cms.IBlobStorage.providedBy(ptah.cms.blobStorage))

    def test_blob_create(self):
        import ptah.cms

        blob = ptah.cms.blobStorage.create()
        self.assertTrue(ptah.cms.IBlob.providedBy(blob))
        self.assertEqual(blob.read(), None)

    def test_blob_metadata(self):
        import ptah.cms

        blob = ptah.cms.blobStorage.add(
            StringIO('blob data'), filename='test.txt', mimetype='text/plain')

        self.assertEqual(blob.filename, 'test.txt')
        self.assertEqual(blob.mimetype, 'text/plain')

    def test_blob_info(self):
        import ptah.cms
        blob = ptah.cms.blobStorage.add(
            StringIO('blob data'), filename='test.txt', mimetype='text/plain')

        info = blob.info()
        self.assertEqual(info['__uri__'], blob.__uri__)
        self.assertEqual(info['filename'], 'test.txt')
        self.assertEqual(info['mimetype'], 'text/plain')

    def test_blob_resolver(self):
        import ptah

        blob = ptah.cms.blobStorage.add(StringIO('blob data'))

        blob_uri = blob.__uri__

        blob = ptah.resolve(blob_uri)

        self.assertEqual(blob.__uri__, blob_uri)
        self.assertEqual(blob.read(), 'blob data')

    def test_blob_with_parent(self):
        import ptah

        class MyContent(ptah.cms.Node):
            __name__ = ''
            __mapper_args__ = {'polymorphic_identity': 'mycontent'}
            __uri_factory__ = ptah.UriFactory('test')

        content = MyContent()
        content_uri = content.__uri__
        ptah.cms.Session.add(content)

        blob_uri = ptah.cms.blobStorage.add(
            StringIO('blob data'), content).__uri__
        transaction.commit()

        blob = ptah.resolve(blob_uri)
        self.assertEqual(blob.__parent_ref__.__uri__, content_uri)

        blob = ptah.cms.blobStorage.getByParent(content_uri)
        self.assertEqual(blob.__uri__, blob_uri)

    def test_blob_write(self):
        import ptah

        blob_uri = ptah.cms.blobStorage.add(StringIO('blob data')).__uri__
        blob = ptah.resolve(blob_uri)
        blob.write('new data')
        transaction.commit()

        blob = ptah.resolve(blob_uri)
        self.assertEqual(blob.read(), 'new data')

    def test_blob_rest_data(self):
        import ptah.cms
        from ptah.cms.rest import blobData
        from pyramid.testing import DummyRequest

        blob = ptah.cms.blobStorage.add(
            StringIO('blob data'), filename='test.txt', mimetype='text/plain')

        request = DummyRequest()

        response = blobData(blob, request)
        self.assertEqual(response.body, 'blob data')
        self.assertEqual(
            response.headerlist,
            [('Content-Disposition', 'filename="test.txt"'),
             ('Content-Length', '9')])



""" test for static assets api """
import os.path
from ptah import view
from ptah.view.base import View
from ptah.view.resources import StaticView

from base import Base

abspath1, pkg1 = view.path('ptah.view.tests:static/dir1')
abspath2, pkg2 = view.path('ptah.view.tests:static/dir2')


class TestStaticManagement(Base):

    def _setup_ptah(self):
        pass

    def test_static_registration_errors(self):
        self.assertRaises(
            ValueError, view.static,
            'tests', 'ptah.view.tests:static/unknown---asdad')

        self.assertRaises(
            ValueError, view.static,
            'tests', 'ptah.view.tests:static/dir1/style.css')

    def test_static_register(self):
        view.static('tests', 'ptah.view.tests:static/dir1')
        self._init_ptah()

        request = self._makeRequest()
        self._setRequest(request)

        self.assertEquals(
            view.static_url(request, 'tests', 'styles.css'),
            'http://localhost:8080/static/tests/styles.css')

        self.assertEquals(
            view.static_url(request, 'tests'),
            'http://localhost:8080/static/tests')

    def test_static_register_url(self):
        from ptah.view import resources
        resources.STATIC.url = 'http://ptah.org/static'

        view.static('testsurl', 'ptah.view.tests:static/dir1')
        self._init_ptah()

        self.assertEquals(
            view.static_url(self.request, 'tests', 'styles.css'),
            'http://ptah.org/static/tests/styles.css')

    def test_static_buildtree(self):
        from ptah.view.resources import buildTree

        abspath, pkg = view.path('ptah.view.tests:static/dir2')
        self.assertEqual(buildTree(abspath),
                         {'style.css': 1, 'text.txt': 1})

        # do not include not allowed files
        abspath, pkg = view.path('ptah.view.tests:static/dir2')
        self.assertTrue('~test.html' not in buildTree(abspath))

        # subtrees
        abspath, pkg = view.path('ptah.view.tests:static')
        self.assertEqual(buildTree(abspath),
                         {os.path.join('dir1','style.css'): 1,
                          os.path.join('dir1','subdir','text.txt'): 1,
                          os.path.join('dir1','text2.txt'): 1,
                          os.path.join('dir2','style.css'): 1,
                          os.path.join('dir2','text.txt'): 1})

    def test_base_static_url(self):
        view.static('tests2', 'ptah.view.tests:static/dir1')
        self._init_ptah()

        request = self._makeRequest()

        base = View(None, request)

        self.assertEquals(
            base.static_url('tests2', 'styles.css'),
            'http://localhost:8080/static/tests2/styles.css')


class TestStaticView(Base):

    def test_resource_empty_pathinfo(self):
        environ = self._makeEnviron(PATH_INFO='')
        request = self._makeRequest(environ)
        inst = StaticView(abspath1, 'static/test')

        response = inst(None, request)
        self.assertTrue(response.status == '404 Not Found')

    def test_resource_doesnt_exist(self):
        environ = self._makeEnviron(PATH_INFO='/static/test/notthere')
        request = self._makeRequest(environ)

        inst = StaticView(abspath1, 'static/test')

        response = inst(None, request)
        self.assertTrue(response.status == '404 Not Found')

    def test_resource_file(self):
        environ = self._makeEnviron(PATH_INFO='/static/test/style.css')
        request = self._makeRequest(environ)
        inst = StaticView(abspath1, 'static/test')

        response = inst(None, request)
        self.assertTrue(response.status == '200 OK')
        self.assertTrue('/* CSS styles file 1 */' in response.body)

    def test_resource_file_from_subdir(self):
        environ = self._makeEnviron(PATH_INFO='/static/test/subdir/text.txt')
        request = self._makeRequest(environ)
        inst = StaticView(abspath1, 'static/test')

        response = inst(None, request)
        self.assertTrue(response.status == '200 OK')
        self.assertTrue('test text 1' in response.body)

    def test_resource_file_with_cache(self):
        environ = self._makeEnviron(PATH_INFO='/static/test/style.css')
        request = self._makeRequest(environ)
        inst = StaticView(abspath1, 'static/test')

        response = inst(None, request)
        self.assertIn('max-age=0', response.headers['Cache-Control'])

        from ptah.view import resources
        resources.STATIC.cache_max_age = 360

        response = inst(None, request)
        self.assertTrue(response.headers['Cache-Control'],
                        'public, max-age=360')

    def test_resource_layers_bypass_to_parent(self):
        # if not exist in upper layer tring to get from lower layer
        environ = self._makeEnviron(PATH_INFO='/static/test/text2.txt')
        request = self._makeRequest(environ)
        inst = StaticView(abspath1, 'static/test')

        response = inst(None, request)
        self.assertTrue(response.status == '200 OK')
        self.assertTrue('text2 test 1' in response.body)

    def test_resource_layers_bypass_to_parent_subdir(self):
        # if not exist in upper layer tring to get from lower layer
        environ = self._makeEnviron(PATH_INFO='/static/test/subdir/text.txt')
        request = self._makeRequest(environ)
        inst = StaticView(abspath1, 'static/test')

        response = inst(None, request)
        self.assertTrue(response.status == '200 OK')
        self.assertTrue('test text 1' in response.body)
